#!/usr/bin/env python

"""
This module provides classes to perform fitting of structures.
"""

from __future__ import division

__author__ = "Stephen Dacek, William Davidson Richards, Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Stephen Dacek"
__email__ = "sdacek@mit.edu"
__status__ = "Production"
__date__ = "Dec 3, 2012"

import numpy as np
import itertools
import abc

from pymatgen.serializers.json_coders import MSONable
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.core.composition import Composition
from pymatgen.optimization.linear_assignment import LinearAssignment
from pymatgen.util.coord_utils import get_points_in_sphere_pbc, \
    pbc_shortest_vectors


class AbstractComparator(MSONable):
    """
    Abstract Comparator class. A Comparator defines how sites are compared in
    a structure.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def are_equal(self, sp1, sp2):
        """
        Defines how the species of two sites are considered equal. For
        example, one can consider sites to have the same species only when
        the species are exactly the same, i.e., Fe2+ matches Fe2+ but not
        Fe3+. Or one can define that only the element matters,
        and all oxidation state information are ignored.

        Args:
            sp1:
                First species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.
            sp2:
                Second species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.

        Returns:
            Boolean indicating whether species are considered equal.
        """
        return

    @abc.abstractmethod
    def get_structure_hash(self, structure):
        """
        Defines a hash for structures. This allows structures to be grouped
        efficiently for comparison. For example, in exact matching,
        you should only try to perform matching if structures have the same
        reduced formula (structures with different formulas can't possibly
        match). So the reduced_formula is a good hash. The hash function
        should be relatively fast to compute relative to the actual matching.

        Args:
            structure:
                A structure

        Returns:
            A hashable object. Examples can be string formulas, etc.
        """
        return

    @classmethod
    def from_dict(cls, d):
        for trans_modules in ['structure_matcher']:
            mod = __import__('pymatgen.analysis.' + trans_modules,
                             globals(), locals(), [d['@class']], -1)
            if hasattr(mod, d['@class']):
                trans = getattr(mod, d['@class'])
                return trans()
        raise ValueError("Invalid Comparator dict")

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__}


class SpeciesComparator(AbstractComparator):
    """
    A Comparator that matches species exactly. The default used in
    StructureMatcher.
    """

    def are_equal(self, sp1, sp2):
        """
        True if species are exactly the same, i.e., Fe2+ == Fe2+ but not Fe3+.

        Args:
            sp1:
                First species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.
            sp2:
                Second species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.

        Returns:
            Boolean indicating whether species are equal.
        """
        return sp1 == sp2

    def get_structure_hash(self, structure):
        """
        Hash for structure.

        Args:
            structure:
                A structure

        Returns:
            Reduced formula for the structure is used as a hash for the
            SpeciesComparator.
        """
        return structure.composition.reduced_formula


class ElementComparator(AbstractComparator):
    """
    A Comparator that matches elements. i.e. oxidation states are
    ignored.
    """

    def are_equal(self, sp1, sp2):
        """
        True if element:amounts are exactly the same, i.e.,
        oxidation state is not considered.

        Args:
            sp1:
                First species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.
            sp2:
                Second species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.

        Returns:
            Boolean indicating whether species are the same based on element
            and amounts.
        """
        comp1 = Composition(sp1)
        comp2 = Composition(sp2)
        return comp1.get_el_amt_dict() == comp2.get_el_amt_dict()

    def get_structure_hash(self, structure):
        """
        Hash for structure.

        Args:
            structure:
                A structure

        Returns:
            Reduced formula for the structure is used as a hash for the
            SpeciesComparator.
        """
        return structure.composition.reduced_formula


class FrameworkComparator(AbstractComparator):
    """
    A Comparator that matches sites, regardless of species.
    """

    def are_equal(self, sp1, sp2):
        """
        True if there are atoms on both sites.

        Args:
            sp1:
                First species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.
            sp2:
                Second species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.

        Returns:
            True always
        """
        return True

    def get_structure_hash(self, structure):
        """
        Hash for structure.

        Args:
            structure:
                A structure

        Returns:
            Number of atoms is a good hash for simple framework matching.
        """
        return len(structure)


class OrderDisorderElementComparator(AbstractComparator):
    """
    A Comparator that matches sites, regardless of species.
    """

    def are_equal(self, sp1, sp2):
        """
        True if there is some overlap in composition between the species

        Args:
            sp1:
                First species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.
            sp2:
                Second species. A dict of {specie/element: amt} as per the
                definition in Site and PeriodicSite.

        Returns:
            True always
        """
        set1 = set(sp1.element_composition.keys())
        set2 = set(sp2.element_composition.keys())
        if set1.intersection(set2):
            return True
        return False

    def get_structure_hash(self, structure):
        """
        Hash for structure.

        Args:
            structure:
                A structure

        Returns:
            TODO No good hash yet
        """
        return 1


class StructureMatcher(MSONable):
    """
    Class to match structures by similarity.

    Algorithm:

    1. Given two structures: s1 and s2
    2. Optional: Reduce to primitive cells.
    3. If the number of sites do not match, return False
    4. Reduce to s1 and s2 to Niggli Cells
    5. Optional: Scale s1 and s2 to same volume.
    6. Optional: Remove oxidation states associated with sites
    7. Find all possible lattice vectors for s2 within shell of ltol.
    8. For s1, translate an atom in the smallest set to the origin
    9. For s2: find all valid lattices from permutations of the list
       of lattice vectors (invalid if: det(Lattice Matrix) < half
       volume of original s2 lattice)
    10. For each valid lattice:

        a. If the lattice angles of are within tolerance of s1,
           basis change s2 into new lattice.
        b. For each atom in the smallest set of s2:

            i. Translate to origin and compare fractional sites in
            structure within a fractional tolerance.
            ii. If true:

                ia. Convert both lattices to cartesian and place
                both structures on an average lattice
                ib. Compute and return the average and max rms
                displacement between the two structures normalized
                by the average free length per atom

                if fit function called:
                    if normalized max rms displacement is less than
                    stol. Return True

                if get_rms_dist function called:
                    if normalized average rms displacement is less
                    than the stored rms displacement, store and
                    continue. (This function will search all possible
                    lattices for the smallest average rms displacement
                    between the two structures)

    """

    def __init__(self, ltol=0.2, stol=0.3, angle_tol=5, primitive_cell=True,
                 scale=True, attempt_supercell=False, allow_subset=False,
                 comparator=SpeciesComparator()):
        """
        Args:
            ltol:
                Fractional length tolerance. Default is 0.2.
            stol:
                Site tolerance. Defined as the fraction of the
                average free length per atom := ( V / Nsites ) ** (1/3)
                Default is 0.3.
            angle_tol:
                Angle tolerance in degrees. Default is 5 degrees.
            primitive_cell:
                If true: input structures will be reduced to primitive
                cells prior to matching. Default to True.
            scale:
                Input structures are scaled to equivalent volume if true;
                For exact matching, set to False.
            attempt_supercell:
                If set to True and number of sites in cells differ
                after a primitive cell reduction (divisible by an integer)
                attempts to generate a supercell transformation of the
                smaller cell which is equivalent to the larger structure.
            allow_subset:
                Allow one structure to match to the subset of another
                structure. Eg. Matching of an ordered structure onto a
                disordered one, or matching a delithiated to a lithiated
                structure. This option cannot be combined with
                attempt_supercell, or with structure grouping.
            comparator:
                A comparator object implementing an equals method that declares
                declaring equivalency of sites. Default is
                SpeciesComparator, which implies rigid species
                mapping, i.e., Fe2+ only matches Fe2+ and not Fe3+.

                Other comparators are provided, e.g., ElementComparator which
                matches only the elements and not the species.

                The reason why a comparator object is used instead of
                supplying a comparison function is that it is not possible to
                pickle a function, which makes it otherwise difficult to use
                StructureMatcher with Python's multiprocessing.
        """
        self.ltol = ltol
        self.stol = stol
        self.angle_tol = angle_tol
        self._comparator = comparator
        self._primitive_cell = primitive_cell
        self._scale = scale
        self._supercell = attempt_supercell
        self._subset = allow_subset
        if attempt_supercell and allow_subset:
            raise ValueError("attempt_supercell cannot be combined with "
                             "allow_subset")

    def _get_lattices(self, target_s, s, supercell_size=1):
        """
        Yields lattices for s with lengths and angles close to the
        lattice of target_s. If supercell_size is specified, the
        returned lattice will have that number of primitive cells
        in it

        Args:
            s, target_s: Structure objects
        """
        t_l, t_a = target_s.lattice.lengths_and_angles
        r = (1 + self.ltol) * max(t_l)
        fpts, dists, i = get_points_in_sphere_pbc(
            lattice=s.lattice, frac_points=[[0, 0, 0]], center=[0, 0, 0],
            r=r).T
        #get possible vectors for a, b, and c
        new_v = []
        for l in t_l:
            max_r = (1 + self.ltol) * l
            min_r = (1 - self.ltol) * l
            vi = fpts[np.where((dists < max_r) & (dists > min_r))]
            if len(vi) == 0:
                return
            cart_vi = np.dot(np.array([i for i in vi]), s.lattice.matrix)
            new_v.append(cart_vi)

        #The vectors are broadcast into a 5-D array containing
        #all permutations of the entries in new_v[0], new_v[1], new_v[2]
        #Produces the same result as three nested loops over the
        #same variables and calculating determinants individually
        bfl = (np.array(new_v[0])[None, None, :, None, :] *
               np.array([1, 0, 0])[None, None, None, :, None] +
               np.array(new_v[1])[None, :, None, None, :] *
               np.array([0, 1, 0])[None, None, None, :, None] +
               np.array(new_v[2])[:, None, None, None, :] *
               np.array([0, 0, 1])[None, None, None, :, None])

        #Compute volume of each lattice
        vol = np.abs(np.sum(bfl[:, :, :, 0, :] *
                            np.cross(bfl[:, :, :, 1, :],
                                     bfl[:, :, :, 2, :]), 3))
        #valid lattices must not change volume
        min_vol = s.volume * 0.999 * supercell_size
        max_vol = s.volume * 1.001 * supercell_size
        bfl = bfl[np.where((vol > min_vol) & (vol < max_vol))]
        if len(bfl) == 0:
            return

        #compute angles
        lengths = np.sum(bfl ** 2, axis=2) ** 0.5
        angles = np.zeros((len(bfl), 3), float)
        for i in xrange(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            angles[:, i] = \
                np.sum(bfl[:, j, :] * bfl[:, k, :], 1) \
                / (lengths[:, j] * lengths[:, k])
        angles = np.arccos(angles) * 180. / np.pi
        #Check angles are within tolerance
        valid_angles = np.where(np.all(np.abs(angles - t_a) <
                                       self.angle_tol, axis=1))
        for lat in bfl[valid_angles]:
            nl = Lattice(lat)
            yield nl

    def _cmp_fractional_struct(self, s1, s2, frac_tol, mask):
        #compares the fractional coordinates
        mask_val = 3 * len(s1)
        dist = s2[:, None] - s1[None, :]
        dist = abs(dist - np.round(dist))
        dist[np.where(dist > frac_tol[None, None, :])] = mask_val
        cost = np.sum(dist, axis=-1)
        cost[mask] = mask_val
        if np.max(np.min(cost, axis=1)) >= mask_val:
            return False
        if self._subset:
            n = len(s1)
            square_cost = np.zeros((n, n))
            square_cost[:cost.shape[0], :cost.shape[1]] = cost
            cost = square_cost
        lin = LinearAssignment(cost)
        if lin.min_cost >= mask_val:
            return False
        return True

    def _cart_dists(self, s1, s2, l1, l2, mask):
        """
        Finds the cartesian distances normalized by (V/Natom) ^ 1/3
        between s1 and s2 on the average lattice of l1 and l2
        s1 and s2 are lists of fractional coordinates. Minimizes the
        RMS distance of the matching with an additional translation
        (but doesn't change the mapping)
        returns distances, fractional_translation vector
        """
        #create the average lattice
        avg_params = (np.array(l1.lengths_and_angles) +
                      np.array(l2.lengths_and_angles)) / 2
        avg_lattice = Lattice.from_lengths_and_angles(*avg_params)
        norm_length = (avg_lattice.volume / len(s1)) ** (1 / 3)
        mask_val = 1e20 * norm_length * self.stol

        all_d_2 = np.zeros([len(s1), len(s1)])
        vec_matrix = np.zeros([len(s1), len(s1), 3])

        vecs = pbc_shortest_vectors(avg_lattice, s2, s1)
        vec_matrix[:len(s2)] = vecs
        vec_matrix[mask] = mask_val
        d_2 = (np.sum(vecs ** 2, axis=-1))
        all_d_2[:len(s2)] = d_2
        all_d_2[mask] = mask_val

        lin = LinearAssignment(all_d_2)
        inds = np.arange(len(s2))
        shortest_vecs = vec_matrix[inds, lin.solution[:len(s2)], :]
        translation = np.average(shortest_vecs, axis=0)
        f_translation = avg_lattice.get_fractional_coords(translation)
        shortest_distances = np.sum((shortest_vecs - translation) ** 2,
                                    -1) ** 0.5
        return shortest_distances / norm_length, f_translation

    def fit(self, struct1, struct2):
        """
        Fit two structures.

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            True or False.
        """

        match = self._find_match(struct1, struct2, break_on_match=True)

        if match is None:
            return False
        else:
            return match[0] <= self.stol

    def get_rms_dist(self, struct1, struct2):
        """
        Calculate RMS displacement between two structures

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            rms displacement normalized by (Vol / nsites) ** (1/3)
            and maximum distance between paired sites. If no matching
            lattice is found None is returned.
        """
        match = self._find_match(struct1, struct2, break_on_match=False,
                                 use_rms=True)
        if match is None:
            return None
        else:
            return match[0], max(match[1])

    def _find_match(self, struct1, struct2, break_on_match=False,
                    use_rms=False, niggli=True):
        """
        Finds the best match between two structures.
        Typically, 'best' is determined by minimax cartesian distance
        on the average lattice

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure
            break_on_match:
                If true, breaks once the max distance is below
                the stol (RMS distance if use_rms is true)
            use_rms:
                If True, finds the match that minimizes
                RMS instead of minimax
            niggli:
                whether to compute the niggli cells of the input
                structures

        Returns:
            the value, distances, s2 lattice, and s2 translation
            vector for the best match
        """
        struct1 = Structure.from_sites(struct1.sites)
        struct2 = Structure.from_sites(struct2.sites)

        if self._comparator.get_structure_hash(struct1) != \
                self._comparator.get_structure_hash(struct2) \
            and not self._subset:
            return None

        #primitive cell transformation
        if self._primitive_cell and struct1.num_sites != struct2.num_sites:
            struct1 = struct1.get_primitive_structure()
            struct2 = struct2.get_primitive_structure()

        if self._supercell:
            #force struct1 to be the larger one
            if struct2.num_sites > struct1.num_sites:
                struct2, struct1 = struct1, struct2
            fu = struct1.num_sites // struct2.num_sites
        else:
            fu = 1

        # Check that the number of sites is correct
        if self._subset:
            if struct2.num_sites > struct1.num_sites:
                struct2, struct1 = struct1, struct2
                #can't do the check until we group with the comparator
        else:
            if struct1.num_sites != struct2.num_sites * fu:
                return None

        # Get niggli reduced cells. Though technically not necessary, this
        # minimizes cell lengths and speeds up the matching of skewed
        # cells considerably.
        if niggli:
            struct1 = struct1.get_reduced_structure(reduction_algo="niggli")
            struct2 = struct2.get_reduced_structure(reduction_algo="niggli")

        nl1 = struct1.lattice
        nl2 = struct2.lattice

        #rescale lattice to same volume
        if self._scale:
            ratio = (fu * nl2.volume / nl1.volume) ** (1 / 6)
            nl1 = Lattice(nl1.matrix * ratio)
            struct1.modify_lattice(nl1)
            nl2 = Lattice(nl2.matrix / ratio)
            struct2.modify_lattice(nl2)

        #fractional tolerance of atomic positions (2x for initial fitting)
        normalization = ((2 * struct2.num_sites) /
                         (struct1.volume + struct2.volume)) ** (1 / 3)
        frac_tol = np.array(struct1.lattice.reciprocal_lattice.abc) * \
                   self.stol / ((1 - self.ltol) * np.pi) / normalization

        #make array mask
        mask = np.zeros((len(struct2) * fu, len(struct1)), dtype=np.bool)
        i = 0
        for site2 in struct2:
            for repeat in range(fu):
                for j, site1 in enumerate(struct1):
                    mask[i, j] = not self._comparator.are_equal(
                        site2.species_and_occu, site1.species_and_occu)
                i += 1

        #print LinearAssignment(mask).min_cost
        #find the best sites for the translation vector
        num_s1_invalid_matches = np.sum(mask, axis=1)
        s2_translation_index = np.argmax(num_s1_invalid_matches)
        s1_translation_indices = np.argwhere(
            mask[s2_translation_index] == 0).flatten()

        s1fc = np.array(struct1.frac_coords)
        s2cc = np.array(struct2.cart_coords)

        best_match = None
        for nl in self._get_lattices(struct1, struct2, fu):
            #if supercell needs to be created, update s2_cart
            if self._supercell and fu > 1:
                scale_matrix = np.round(np.dot(nl.matrix, nl2.inv_matrix))
                supercell = struct2.copy()
                supercell.make_supercell(scale_matrix.astype('int'))
                s2fc = np.array(supercell.frac_coords)
            else:
                s2fc = nl.get_fractional_coords(s2cc)
                #loop over possible translations
            for s1i in s1_translation_indices:
                translation = s1fc[s1i] - s2fc[s2_translation_index]
                t_s2fc = s2fc + translation
                if self._cmp_fractional_struct(s1fc, t_s2fc, frac_tol, mask):
                    distances, t = self._cart_dists(s1fc, t_s2fc, nl, nl1,
                                                    mask)
                    if use_rms:
                        val = np.linalg.norm(distances) / len(distances) ** \
                              0.5
                    else:
                        val = max(distances)
                    if best_match is None or val < best_match[0]:
                        total_translation = translation + t
                        total_translation -= np.round(total_translation)
                        best_match = val, distances, nl, total_translation
                    if break_on_match and val < self.stol:
                        return best_match
        if best_match and best_match[0] < self.stol:
            return best_match

    def find_indexes(self, s_list, group_list):
        """
        Given a list of structures, return list of indices where each
        structure appears in group_list.

        Args:
            s_list:
                list of structures to check
            group_list:
                list to find structures in
        """
        inds = [-1] * len(s_list)
        for j in range(len(s_list)):
            for i in range(len(group_list)):
                if s_list[j] in group_list[i]:
                    inds[j] = i
                    break
        return inds

    def group_structures(self, s_list):
        """
        Given a list of structures, use fit to group
        them by structural equality.

        Args:
            s_list:
                List of structures to be grouped

        Returns:
            A list of lists of matched structures
            Assumption: if s1=s2 and s2=s3, then s1=s3
            This may not be true for small tolerances.
        """
        if self._subset:
            raise ValueError("allow_subset cannot be used with"
                             " group_structures")
            #Use structure hash to pre-group structures.
        sorted_s_list = sorted(s_list, key=self._comparator.get_structure_hash)
        all_groups = []

        #For each pre-grouped list of structures, perform actual matching.
        for k, g in itertools.groupby(sorted_s_list,
                                      key=self._comparator.get_structure_hash):
            g = list(g)
            group_list = [[g[0]]]
            for i, j in itertools.combinations(range(len(g)), 2):
                s1_ind, s2_ind = self.find_indexes([g[i], g[j]], group_list)
                if s2_ind == -1 and self.fit(g[i], g[j]):
                    group_list[s1_ind].append(g[j])
                elif (j - i) == 1 and s2_ind == -1:
                    group_list.append([g[j]])
            all_groups.extend(group_list)
        return all_groups

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "comparator": self._comparator.to_dict,
                "stol": self.stol,
                "ltol": self.ltol,
                "angle_tol": self.angle_tol,
                "primitive_cell": self._primitive_cell,
                "scale": self._scale}

    @classmethod
    def from_dict(cls, d):
        return StructureMatcher(
            ltol=d["ltol"], stol=d["stol"], angle_tol=d["angle_tol"],
            primitive_cell=d["primitive_cell"], scale=d["scale"],
            comparator=AbstractComparator.from_dict(d["comparator"]))

    def get_minimax_rms_anonymous(self, struct1, struct2):
        """
        Performs an anonymous fitting, which allows distinct species in one
        structure to map to another. E.g., to compare if the Li2O and Na2O
        structures are similar.

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            (minimax_rms, min_mapping)
            min_rms is the minimum of the max rms calculated, and min_mapping
            is the corresponding minimal species mapping that would map
            struct1 to struct2. (None, None) is returned if the minimax_rms
            exceeds the threshold.
        """
        sp1 = list(set(struct1.species_and_occu))
        sp2 = list(set(struct2.species_and_occu))

        if len(sp1) != len(sp2):
            return None, None

        latt1 = struct1.lattice
        fcoords1 = struct1.frac_coords
        min_rms = float("inf")
        min_mapping = None
        for perm in itertools.permutations(sp2):
            sp_mapping = dict(zip(sp1, perm))
            mapped_sp = [sp_mapping[site.species_and_occu] for site in struct1]
            transformed_structure = Structure(latt1, mapped_sp, fcoords1)
            rms = self.get_rms_dist(transformed_structure, struct2)
            if rms is not None:
                if min_rms > rms[1]:
                    min_rms = rms[1]
                    min_mapping = {k: v for k, v in sp_mapping.items()
                                   if k != v}
        if min_mapping is None:
            return None, None
        else:
            return min_rms, min_mapping

    def fit_with_electronegativity(self, struct1, struct2):
        """
        Performs an anonymous fitting, which allows distinct species in one
        structure to map to another. E.g., to compare if the Li2O and Na2O
        structures are similar. If multiple substitutions are within tolerance
        this will return the one which minimizes the difference in electronegativity
        between the matches species.

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            min_mapping
            min_rms is the minimum of the max rms calculated, and min_mapping
            is the corresponding minimal species mapping that would map
            struct1 to struct2. None is returned if no fit is found.
        """

        sp1 = list(set(struct1.species_and_occu))
        sp2 = list(set(struct2.species_and_occu))
        if len(sp1) != len(sp2):
            return None, None

        latt1 = struct1.lattice
        fcoords1 = struct1.frac_coords
        min_X_diff = np.inf
        min_mapping = None
        for perm in itertools.permutations(sp2):
            sp_mapping = dict(zip(sp1, perm))
            mapped_sp = [sp_mapping[site.species_and_occu] for site in struct1]
            transformed_structure = Structure(latt1, mapped_sp, fcoords1)
            if self.fit(transformed_structure, struct2):
                #Calculate electronegativity difference
                X_diff = np.average(
                    [(host_sp.elements[0].X - map_sp.elements[0].X) *
                     struct1.composition.get(host_sp.elements[0]) for
                     host_sp, map_sp in sp_mapping.iteritems()])

                if min_X_diff is 0:
                    return {}

                if min_X_diff > X_diff:
                    min_X_diff = X_diff
                    min_mapping = {k: v for k, v in sp_mapping.items()
                                   if k != v}
        if min_mapping is None:
            return None
        else:
            return min_mapping

    def fit_anonymous_all_mapping(self, struct1, struct2):
        """
        Performs an anonymous fitting, which allows distinct species in one
        structure to map to another. Returns a dictionary of species
        substitutions that are within tolerance

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            (mappings)
            mappings is a list of possible species mappings that
            would map struct1 to struct2.
        """
        sp1 = list(set(struct1.species_and_occu))
        sp2 = list(set(struct2.species_and_occu))

        if len(sp1) != len(sp2):
            return None

        latt1 = struct1.lattice
        fcoords1 = struct1.frac_coords
        mappings = []
        for perm in itertools.permutations(sp2):
            sp_mapping = dict(zip(sp1, perm))
            mapped_sp = [sp_mapping[site.species_and_occu] for site in struct1]
            transformed_structure = Structure(latt1, mapped_sp, fcoords1)
            rms = self.get_rms_dist(transformed_structure, struct2)
            if rms is not None:

                possible_mapping = {k: v for k, v in sp_mapping.items()}

                if rms[1] < self.stol:
                    #check if mapping already found
                    for k, v in possible_mapping.iteritems():
                        if {k: v} not in mappings:
                            mappings.append({k: v})
        if not mappings:
            return None
        else:
            return mappings

    def fit_anonymous(self, struct1, struct2):
        """
        Performs an anonymous fitting, which allows distinct species in one
        structure to map to another. E.g., to compare if the Li2O and Na2O
        structures are similar.

        Args:
            struct1:
                1st structure
            struct2:
                2nd structure

        Returns:
            A minimal species mapping that would map struct1 to struct2 in
            terms of structure similarity, or None if no fit is found. For
            example, to map the cubic Li2O to cubic Na2O,
            we need a Li->Na mapping. This method will return
            [({Element("Li"): 1}, {Element("Na"): 1})]. Since O is the same
            in both structures, there is no O to O mapping required.
            Note that the return form is a list of pairs of species and
            occupancy dicts. This complicated return for is necessary because
            species and occupancy dicts are non-hashable.
        """
        min_rms, min_mapping = self.get_minimax_rms_anonymous(struct1, struct2)

        if min_rms is None or min_rms > self.stol:
            return None
        else:
            return min_mapping

    def get_s2_like_s1(self, struct1, struct2):
        """
        performs transformations on struct2 to put it in a basis similar to
        struct1 (without changing any of the inter-site distances)
        """
        if self._primitive_cell:
            raise ValueError("get_s2_like_s1 cannot be used with the primitive"
                             " cell option")
        if self._supercell and struct2.num_sites > struct1.num_sites:
            raise ValueError("The non-supercell must be put onto the basis"
                             " of the supercell, not the other way around")
        if self._subset and struct2.num_sites > struct1.num_sites:
            raise ValueError("The smaller structure must be transformed onto"
                             " the larger one, not the other way around")
        match = self._find_match(struct1, struct2, break_on_match=False,
                                 use_rms=True, niggli=False)
        if match is None:
            return None
        scale_matrix = np.round(
            np.dot(match[2].matrix, struct2.lattice.inv_matrix)).astype('int')
        temp = struct2.copy()
        temp.make_supercell(scale_matrix)
        fc = temp.frac_coords + match[3]
        return Structure(temp.lattice, temp.species_and_occu, fc,
                         to_unit_cell=True)



"""
Factory functions producing ABINIT workflows. Entry points for client code
(high-level interface)
"""
from __future__ import division, print_function

import os

from pymatgen.io.abinitio.abiobjects import (Smearing, KSampling, Screening,
    SelfEnergy, ExcHamiltonian)

from pymatgen.io.abinitio.strategies import (ScfStrategy, NscfStrategy,
    ScreeningStrategy, SelfEnergyStrategy, MDFBSE_Strategy)

from pymatgen.io.abinitio.workflow import (PseudoIterativeConvergence, 
    PseudoConvergence, BandStructure, GW_Workflow, BSEMDF_Workflow)

__author__ = "Matteo Giantomassi"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"
__email__ = "gmatteo at gmail.com"


################################################################################

class PPConvergenceFactory(object):
    """
    Factory object that constructs workflows for analyzing the converge of
    pseudopotentials.
    """
    def work_for_pseudo(self, workdir, pseudo, ecut_range, 
                        runmode="sequential", toldfe=1.e-8,
                        atols_mev=(10, 1, 0.1), spin_mode="polarized",
                        acell=(8, 9, 10), smearing="fermi_dirac:0.1 eV",):
        """
        Return a Work object given the pseudopotential pseudo.

        Args:
            workdir:
                Working directory.
            pseudo:
                Pseudo object.
            ecut_range:
                range of cutoff energies in Ha units.
            runmode:
                Run mode.
            toldfe:
                Tolerance on the total energy (Ha).
            atols_mev:
                Tolerances in meV for accuracy in ["low", "normal", "high"]
            spin_mode:
                Spin polarization.
            acell:
                Length of the real space lattice (Bohr units)
            smearing:
                Smearing technique.
        """
        workdir = os.path.abspath(workdir)

        smearing = Smearing.assmearing(smearing)

        if isinstance(ecut_range, slice):
            workflow = PseudoIterativeConvergence(
                workdir, pseudo, ecut_range, atols_mev, 
                runmode=runmode, toldfe=toldfe, spin_mode=spin_mode, 
                acell=acell, smearing=smearing)

        else:
            workflow = PseudoConvergence(
                workdir, pseudo, ecut_range, atols_mev, 
                runmode=runmode, toldfe=toldfe, spin_mode=spin_mode, 
                acell=acell, smearing=smearing)

        return workflow

################################################################################


def bandstructure(workdir, runmode, structure, pseudos, scf_kppa, nscf_nband,
                  ndivsm, accuracy="normal", spin_mode="polarized",
                  smearing="fermi_dirac:0.1 eV", charge=0.0, scf_algorithm=None,
                  dos_kppa=None, **extra_abivars):
    """
    Returns a Work object that computes that bandstructure of the material.

    Args:
        workdir:
            Working directory.
        runmode:
            `RunMode` instance.
        structure:
            Pymatgen structure.
        pseudos:
            List of `Pseudo` objects.
        scf_kppa:
            Defines the sampling used for the SCF run.
        nscf_nband:
            Number of bands included in the NSCF run.
        ndivs:
            Number of divisions used to sample the smallest segment of the
            k-path.
        accuracy:
            Accuracy of the calculation.
        spin_mode:
            Spin polarization.
        smearing:
            Smearing technique.
        charge:
            Electronic charge added to the unit cell.
        scf_algorithm:
            Algorithm used for solving of the SCF cycle.
        dos_kppa:
            Defines the k-point sampling used for the computation of the DOS 
            (None if DOS is not wanted).
        **extra_abivars:
            Dictionary with extra variables passed to ABINIT.
    """
    # SCF calculation.
    scf_ksampling = KSampling.automatic_density(structure, scf_kppa, chksymbreak=0)

    scf_strategy = ScfStrategy(structure, pseudos, scf_ksampling,
                               accuracy=accuracy, spin_mode=spin_mode,
                               smearing=smearing, charge=charge,
                               scf_algorithm=scf_algorithm, **extra_abivars)

    # Band structure calculation.
    nscf_ksampling = KSampling.path_from_structure(ndivsm, structure)

    nscf_strategy = NscfStrategy(scf_strategy, nscf_ksampling, nscf_nband, **extra_abivars)

    # DOS calculation.
    dos_strategy = None
    if dos_kppa is not None:
        raise NotImplementedError("DOS must be tested")
        dos_ksampling = KSampling.automatic_density(structure, dos_kppa, chksymbreak=0)
        #dos_ksampling = KSampling.monkhorst(dos_ngkpt, shiftk=dos_shiftk, chksymbreak=0)

        dos_strategy = NscfStrategy(scf_strategy, dos_ksampling, nscf_nband, nscf_solver=None, **extra_abivars)

    return BandStructure(workdir, runmode, scf_strategy, nscf_strategy, dos_strategy=dos_strategy)

################################################################################


#def relaxation(workdir, runmode, structure, pseudos, scf_kppa,
#               accuracy="normal", spin_mode="polarized",
#               smearing="fermi_dirac:0.1 eV", charge=0.0, scf_algorithm=None, **extra_abivars):
#    """
#    Returns a Work object that performs structural relaxations.
#
#    Args:
#        workdir:
#            Working directory.
#        runmode:
#            `RunMode` instance.
#        structure:
#            Pymatgen structure.
#        pseudos:
#            List of `Pseudo` objects.
#        scf_kppa:
#            Defines the sampling used for the SCF run.
#        accuracy:
#            Accuracy of the calculation.
#        spin_mode:
#            Spin polarization.
#        smearing:
#            Smearing technique.
#        charge:
#            Electronic charge added to the unit cell.
#        scf_algorithm:
#            Algorithm used for solving the SCF cycle.
#    """
#    # SCF calculation.
#    scf_ksampling = KSampling.automatic_density(structure, scf_kppa, chksymbreak=0)
#    relax_algo = 
#
#    relax_strategy = RelaxStrategy(structure, pseudos, scf_ksampling, relax_algo, 
#                                   accuracy=accuracy, spin_mode=spin_mode, smearing=smearing, 
#                                   charge=charge, scf_algorithm=scf_algorithm)
#
#    #return Relaxation(workdir, runmode, relax_strategy)

################################################################################

def g0w0_with_ppmodel(workdir, runmode, structure, pseudos, scf_kppa,
                      nscf_nband, ecuteps, ecutsigx, accuracy="normal",
                      spin_mode="polarized", smearing="fermi_dirac:0.1 eV",
                      ppmodel="godby", charge=0.0, scf_algorithm=None,
                      inclvkb=2, scr_nband=None, sigma_nband=None, **extra_abivars):
    """
    Returns a Work object that performs G0W0 calculations for the given the material.

    Args:
        workdir:
            Working directory.
        runmode:
            `RunMode` instance.
        structure:
            Pymatgen structure.
        pseudos:
            List of `Pseudo` objects.
        scf_kppa:
            Defines the sampling used for the SCF run.
        nscf_nband:
            Number of bands included in the NSCF run.
        ecuteps:
            Cutoff energy [Ha] for the screening matrix.
        ecutsigx:
            Cutoff energy [Ha] for the exchange part of the self-energy.
        accuracy:
            Accuracy of the calculation.
        spin_mode:
            Spin polarization.
        smearing:
            Smearing technique.
        ppmodel:
            Plasmonpole technique.
        charge:
            Electronic charge added to the unit cell.
        scf_algorithm:
            Algorithm used for solving of the SCF cycle.
        inclvkb:
            Treatment of the dipole matrix elements (see abinit variable).
        scr_nband:
            Number of bands used to compute the screening (default is nscf_nband)
        sigma_nband:
            Number of bands used to compute the self-energy (default is nscf_nband)
        **extra_abivars
            Dictionary with extra variables passed to ABINIT.
    """
    # TODO: Cannot use istwfk != 1.
    if "istwfk" not in extra_abivars:
        extra_abivars["istwfk"] = "*1"

    scf_ksampling = KSampling.automatic_density(structure, scf_kppa, chksymbreak=0)

    scf_strategy = ScfStrategy(structure, pseudos, scf_ksampling,
                               accuracy=accuracy, spin_mode=spin_mode,
                               smearing=smearing, charge=charge,
                               scf_algorithm=None, **extra_abivars)

    nscf_ksampling = KSampling.automatic_density(structure, 1, chksymbreak=0)

    nscf_strategy = NscfStrategy(scf_strategy, nscf_ksampling, nscf_nband, **extra_abivars)

    if scr_nband is None:
        scr_nband = nscf_nband

    if sigma_nband is None:
        sigma_nband = nscf_nband

    screening = Screening(ecuteps, scr_nband, w_type="RPA", sc_mode="one_shot",
                          freq_mesh=None, hilbert_transform=None, ecutwfn=None,
                          inclvkb=inclvkb)

    self_energy = SelfEnergy("gw", "one_shot", sigma_nband, ecutsigx, screening,
                             ppmodel=ppmodel)

    scr_strategy = ScreeningStrategy(scf_strategy, nscf_strategy, screening, **extra_abivars)

    sigma_strategy = SelfEnergyStrategy(scf_strategy, nscf_strategy, scr_strategy, self_energy,
                                        **extra_abivars)

    return GW_Workflow(workdir, runmode, scf_strategy, nscf_strategy, scr_strategy, sigma_strategy)

################################################################################


def bse_with_mdf(workdir, runmode, structure, pseudos, scf_kppa, nscf_nband, 
                 nscf_ngkpt, nscf_shiftk, ecuteps, bs_loband, soenergy, mdf_epsinf, 
                 accuracy="normal", spin_mode="polarized", smearing="fermi_dirac:0.1 eV",
                 charge=0.0, scf_algorithm=None, **extra_abivars):
    """
    Returns a Work object that performs a GS + NSCF + Bethe-Salpeter calculation.
    The self-energy corrections are approximated with the scissors operator. The screening
    in modeled by the model dielectric function.

    Args:
        workdir:
            Working directory.
        runmode:
            `RunMode` instance.
        structure:
            Pymatgen structure.
        pseudos:
            List of `Pseudo` objects.
        scf_kppa:
            Defines the sampling used for the SCF run.
        nscf_nband:
            Number of bands included in the NSCF run.
        nscf_ngkpt:
            Division of the k-mesh used for the NSCF and the BSE run.
        nscf_shiftk:
            Shifts used for the NSCF and the BSE run.
        ecuteps:
            Cutoff energy [Ha] for the screening matrix.
        bs_loband:
            Index of the first occupied band included the e-h basis set
            (ABINIT convention i.e. first band starts at 1).
        soenergy:
            Scissor energy in Hartree
        mdf_epsinf:
            Value of the macroscopic dielectric function used in expression for the model dielectric function.
        accuracy:
            Accuracy of the calculation.
        spin_mode:
            Spin polarization.
        smearing:
            Smearing technique.
        charge:
            Electronic charge added to the unit cell.
        scf_algorithm:
            Algorithm used for solving of the SCF cycle.
        **extra_abivars
            Dictionary with extra variables passed to ABINIT.
    """
    # TODO: Cannot use istwfk != 1.
    if "istwfk" not in extra_abivars:
        extra_abivars["istwfk"] = "*1"

    # Ground-state strategy.
    scf_ksampling = KSampling.automatic_density(structure, scf_kppa, chksymbreak=0)

    scf_strategy = ScfStrategy(structure, pseudos, scf_ksampling,
                               accuracy=accuracy, spin_mode=spin_mode,
                               smearing=smearing, charge=charge, scf_algorithm=None, **extra_abivars)

    # NSCF calculation on the randomly-shifted k-mesh.
    nscf_ksampling = KSampling.monkhorst(nscf_ngkpt, shiftk=nscf_shiftk, chksymbreak=0)

    nscf_strategy = NscfStrategy(scf_strategy, nscf_ksampling, nscf_nband, **extra_abivars)

    # Strategy for the BSE calculation.
    raise NotImplementedError("")
    # FIXME
    bs_nband = 6
    coulomb_mode = "model_df"
    bs_freq_mesh = [0, 2, 0.1]

    exc_ham = ExcHamiltonian(bs_loband, bs_nband, soenergy, coulomb_mode, ecuteps, bs_freq_mesh, 
                             mdf_epsinf=mdf_epsinf, exc_type="TDA", algo="haydock", with_lf=True, 
                             zcut=None)


    bse_strategy = MDFBSE_Strategy(scf_strategy, nscf_strategy, exc_ham, **extra_abivars)

    return BSEMDF_Workflow(workdir, runmode, scf_strategy, nscf_strategy, bse_strategy)




#!/usr/bin/env python

"""
This module implements symmetry-related structure forms.
"""

from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Mar 9, 2012"

import numpy as np
from pymatgen.core.structure import Structure


class SymmetrizedStructure(Structure):
    """
    This class represents a symmetrized structure, i.e. a structure
    where the spacegroup and symmetry operations are defined. This class is
    typically not called but instead is typically obtained by calling
    pymatgen.symmetry.SymmetryFinder.get_symmetrized_structure.
    
    .. attribute: equivalent_indices
        
        indices of structure grouped by equivalency
    """

    def __init__(self, structure, spacegroup, equivalent_positions):
        """
        Args:
            structure:
                Original structure
            spacegroup:
                An input spacegroup from SymmetryFinder.
            equivalent_positions:
                Equivalent positions from SymmetryFinder.
        """
        Structure.__init__(self, structure.lattice,
                           [site.species_and_occu
                            for site in structure],
                           structure.frac_coords,
                           site_properties=structure.site_properties)
        
        self._spacegroup = spacegroup
        u, inv = np.unique(equivalent_positions, return_inverse = True)
        self.equivalent_indices = [[] for i in xrange(len(u))]
        self._equivalent_sites = [[] for i in xrange(len(u))]
        for i, inv in enumerate(inv):
            self.equivalent_indices[inv].append(i)
            self._equivalent_sites[inv].append(self.sites[i])

    @property
    def equivalent_sites(self):
        """
        All the sites grouped by symmetry equivalence in the form of [[sites
        in group1], [sites in group2], ...]
        """
        return self._equivalent_sites

    def find_equivalent_sites(self, site):
        """
        Finds all symmetrically equivalent sites for a particular site

        Args:
            site:
                A site in the structure

        Returns:
            A list of all symmetrically equivalent sites.
        """
        for sites in self.equivalent_sites:
            if site in sites:
                return sites

        raise ValueError("Site not in structure")



#!/usr/bin/env python
from __future__ import division, print_function

import os 
import abc

from subprocess import Popen, PIPE

from pymatgen.io.abinitio.utils import File


__all__ = [
    "ScriptEditor",
    "ShellLauncher",
]

# =========================================================================== #

class ScriptEditor(object):
    """Simple editor that simplifies the writing of shell scripts"""
    _shell = '/bin/bash'

    def __init__(self):
        self._lines = []

    @property
    def shell(self):
        return self._shell

    def _add(self, text, pre=""):
        if isinstance(text, str):
            self._lines.append(pre+text)
        else:
            self._lines.extend([pre + t for t in text])

    def reset(self):
        """Reset the editor."""
        try:
            del self._lines
        except AttributeError:
            pass

    def shebang(self):
        """Adds the shebang line."""
        self._lines.append('#!' + self.shell)

    def declare_var(self, key, val):
        """Return a lines setting a variable."""
        line = key + '=' + str(val)
        self._add(line)

    def declare_vars(self, d):
        """Declare the variabled defined in d (dict)"""
        for k,v in d.items():
            self.declare_var(k, v)

    def export_envar(self, key, val):
        """Export an environment variable."""
        line = "export " + key + "=" + str(val)
        self._add(line)

    def export_envars(self, env):
        """Export the environment variables contained in the dict env."""
        for k,v in env.items():
            self.export_envar(k, v)

    def add_emptyline(self):
        self._add("", pre="")

    def add_comment(self, comment):
        self._add(comment, pre="# ")

    def load_modules(self, modules):
        for module in modules:
            self.load_module(module)

    def load_module(self, module):
        self._add('module load ' + module)

    def add_line(self, line):
        self._add(line)

    def add_lines(self, lines):
        self._add(lines)

    def get_script_str(self, reset=True):
        """Returns a string with the script and reset the editor if reset is True"""
        s = "\n".join(l for l in self._lines)
        if reset:
            self.reset()
        return s

##########################################################################################

class OMPEnv(dict):
    """
    Dictionary with the OpenMP environment variables"
    see https://computing.llnl.gov/tutorials/openMP/#EnvironmentVariables"
    """
    _KEYS = [
       "OMP_SCHEDULE",
       "OMP_NUM_THREADS",
       "OMP_DYNAMIC",
       "OMP_PROC_BIND",
       "OMP_NESTED",
       "OMP_STACKSIZE",
       "OMP_WAIT_POLICY",
       "OMP_MAX_ACTIVE_LEVELS",
       "OMP_THREAD_LIMIT",
       "OMP_STACKSIZE",
       "OMP_PROC_BIND",
    ]

    def __init__(self, *args, **kwargs):
        """ 
        Constructor method inherited from dictionary:
                                                   
        >>> OMPEnv(OMP_NUM_THREADS=1)
        {'OMP_NUM_THREADS': '1'}
                                                   
        To create an instance from an INI file, use:
           OMPEnv.from_filename(filename)
        """
        self.update(*args, **kwargs)

        err_msg = ""
        for key, value in self.items():
            self[key] = str(value)
            if key not in OMPEnv._KEYS:
                err_msg += "unknown option %s" % key

        if err_msg: 
            raise ValueError(err_msg)

    @staticmethod
    def from_filename(filename, allow_empty=False):
        from ConfigParser import SafeConfigParser, NoOptionError
        parser = SafeConfigParser()
        parser.read(filename)

        obj = OMPEnv()

        # Consistency check. Note that we only check if the option name is correct, 
        # we do not check whether the value is correct or not.
        if "openmp" not in parser.sections():
            if not allow_empty:
                raise ValueError("%s does not contain any [openmp] section" % filename) 
            return obj

        err_msg = ""
        for key in parser.options("openmp"):
            if key.upper() not in OMPEnv._KEYS:
                err_msg += "unknown option %s, maybe a typo" % key

        if err_msg: 
            raise ValueError(err_msg)

        for key in OMPEnv._KEYS:
            try:
                obj[key] = str(parser.get("openmp", key))
            except NoOptionError:
                try:
                    obj[key] = str(parser.get("openmp", key.lower()))
                except NoOptionError:
                    pass

        if not allow_empty and not obj:
            raise ValueError("Refusing to return with an empty dict") 

        return obj

##########################################################################################

class TaskLauncher(object):
    """Abstract class for a task launcher."""
    __metaclass__ = abc.ABCMeta

    @staticmethod
    def from_launcher_type(launcher_type):
        """Returns the subclass from a string giving its type."""
        classes = []
        for cls in TaskLauncher.__subclasses__():
           if cls._type == launcher_type:
               classes.append(cls)
        if len(classes) != 1:
            raise ValueError("Cannot find class from launcher_type %s" % launcher_type)

        return classes[0]

    def __init__(self, path, **kwargs):
        """
        Args:
            path:
                Path to the script file.
            stdin:
                The input file to feed in the executable as the standard input. Mandatory.
            stdout:
                The file into which the standard output is redirected. Default is 'log'.
            stderr:
                The file into which the standard error is redirected. Default is 'stderr'.
            bindir:
                The directory in which to look for binaries. Default is "".
            modules:
                The modules which will be loaded with 'module load'. Default is None
            pre_lines:
                Lines before the main execution. Default is []
            post_lines:
                Lines after the main execution. Default is [].
            exe:
                The name of the binary to be executed 
                (located in bindir or in $PATH if not bindir). Default is abinit.
            mpirun:
                The mpi runner.
                E.g. 'mpirun -n 6'. Default is None.
            mpi_ncpus
                Number of MPI processes (default 1)
            omp_env
                Dictionary with the OMP environment variables. Default is {}
            vars:
                Dictionary {varname:varvalue} with variable declaration.
        """
        self.jobfile    = File(path)
        self.stdin      = kwargs.get("stdin",  "abi.files")
        self.stdout     = kwargs.get("stdout", "abi.log")
        self.stderr     = kwargs.get("stderr", "abi.err")
        self.bindir     = kwargs.get("bindir", "")
        self.exe        = os.path.join(self.bindir, kwargs.get("exe", "abinit"))
        self.mpirun     = kwargs.get("mpirun", "")
        self.mpi_ncpus  = kwargs.get("mpi_ncpus", 1)
        self.omp_env    = kwargs.get("omp_env", {})
        if self.omp_env:
            self.omp_env = OMPEnv(self.omp_env)
        self.vars         = kwargs.get("vars", {})
        self.modules      = kwargs.get("modules", [])
        self.envars       = kwargs.get("envars", {})
        self.pre_lines    = kwargs.get("pre_lines", [])
        self.post_lines   = kwargs.get("post_lines", [])
        self.queue_params = kwargs.get("queue_params", {})

    @property
    def tot_ncpus(self):
        """Total number of CPUs employed"""
        return self.mpi_ncpus * self.omp_ncpus 

    @property
    def has_mpirun(self):
        """True if we are using a mpirunner"""
        return bool(self.mpirun)

    @property
    def has_omp(self):
        """True if we are using OMP threads"""
        return hasattr(self,"omp_env") and bool(getattr(self, "omp_env"))
                                                      
    @property
    def omp_ncpus(self):
        """Number of OMP threads."""
        if self.has_omp:
            return self.omp_env["OMP_NUM_THREADS"]
        else:
            return 1

    def get_script_str(self):
        """Returns a string wth the script."""
        se = ScriptEditor()

        se.shebang()

        # Submission instructions for the queue manager.
        se.add_lines(self.make_rsmheader())

        se.add_comment("Variable declarations")
        vars = self.vars.copy()

        vars.update({
                "EXECUTABLE": self.exe,
                "STDIN"     : self.stdin,
                "STDOUT"    : self.stdout,
                "STDERR"    : self.stderr,
                "MPIRUN"    : self.mpirun,
                "MPI_NCPUS" : self.mpi_ncpus,
               })

        se.declare_vars(vars)

        se.add_comment("Modules")
        se.load_modules(self.modules)

        se.add_comment("Environment")
        se.export_envars(self.envars)

        if self.has_omp:
            se.export_envars(self.omp_env)

        se.add_comment("Commands before execution")
        se.add_lines(self.pre_lines)

        # Execution line
        # TODO: better treatment of mpirun syntax.
        if self.mpirun:
            se.add_line('$MPIRUN -n $MPI_NCPUS $EXECUTABLE < $STDIN > $STDOUT 2> $STDERR')
        else:
            se.add_line('$EXECUTABLE < $STDIN > $STDOUT 2> $STDERR')

        se.add_comment("Commands after execution")
        se.add_lines(self.post_lines)

        return se.get_script_str()

    def write_script(self, overwrite=True):
        """Writes the script file."""
        #if not self.jobfile.exists:
        #    os.makedirs(self.jobfile.dirname)

        #if self.jobfile.exists and not overwrite:
        #    raise ValueError("%s already exists, cannot overwrite" % self.jobfile.path)

        with open(self.jobfile.path, 'w') as f:
            f.write(self.get_script_str())

    # TODO
    #@abc.abstractproperty
    #def has_resource_manager(self):
    #    "True if job submission is handled by a resource manager."
    #    #return self.launcher not in ["shell",]

    @abc.abstractmethod
    def make_rsmheader(self):
        """Return a list of string with the options passed to the resource manager"""

    @abc.abstractmethod
    def launch(self, task, *args, **kwargs):
        """
        Run the script in a subprocess, returns a Popen-like object.

        Args:
            task:
                Task instance. launch should call task.set_status to modify the status 
                of the object. 
                task.set_status(Task.S_SUB) is called before submitting the task
                task.set_status(Task.S_RUN) is called once the task has started to run.
        """

##########################################################################################
# Concrete classes

class ShellLauncher(TaskLauncher):
    _type = "shell"

    def make_rsmheader(self):
        return []

    def launch(self, task, *args, **kwargs):
        self.write_script()

        task.set_status(task.S_SUB)
        process = Popen(("/bin/bash", self.jobfile.path), cwd=self.jobfile.dirname, stderr=PIPE)
        task.set_status(task.S_RUN)

        return process

##########################################################################################

#class SlurmLauncher(TaskLauncher):
#    _type = "slurm"
#
#    def make_rsmheader(self):
#        raise NotImplementedError()
#
#    def launch(self, *args, **kwargs):
#        """
#        Run the script in a subprocess, returns a Popen-like object.
#        This is a very delicate part. Hopefully fireworks will help solve the problem!
#        """
#        raise NotImplementedError()

##########################################################################################


class SimpleResourceManagerError(Exception):
    """Base error class for `SimpleResourceManager."""


class SimpleResourceManager(object):

    Error = SimpleResourceManagerError

    def __init__(self, work, max_ncpus, sleep_time=20, verbose=0):
        """
            Args:
                work:
                    Work instance.
                max_ncpus: 
                    The maximum number of CPUs that can be used.
                sleep_time:
                    Time delay (seconds) before trying to start a new task.
        """
        from pymatgen.io.abinitio.workflow import Workflow
        if not isinstance(work, Workflow):
            work = Workflow.from_task(work)

        self.work = work

        self.max_ncpus = max_ncpus 
        self.sleep_time = sleep_time
        self.verbose = verbose

        for task in self.work:
            if task.tot_ncpus > self.max_ncpus:
                err_msg = "Task %s requires %s CPUs, but max_ncpus is %d" % (
                    repr(task), task.tot_ncpus, max_ncpus)
                raise self.Error(err_msg)

    def run(self, *args, **kwargs):
        """Call the start method of the object contained in work."""
        while True:
            polls = self.work.poll()
            # Fetch the first task that is ready to run
            try:
                task = self.work.fetch_task_to_run()
            except StopIteration:
                break

            if task is None:
                import time
                time.sleep(self.sleep_time)
            else:
                # Check that we don't exceed the number of cpus employed, before starting.
                if self.verbose:
                    print("work polls %s" % polls)
                    print("work status %s" % self.work.get_status())

                if (task.tot_ncpus + self.work.ncpus_reserved <= self.max_ncpus): 
                    if self.verbose: print("Starting task %s" % task)
                    task.start()

        # Wait until all tasks are completed.
        self.work.wait()

        #if any([t.status != t.S_DONE for t in self]):
        #   for task in self:
        #       print([link for link in task._links])
        #       #print([link_stat==task.S_DONE for link_stat in task.links_status])
        #       msg = "Deadlock, likely due to task dependencies: status %s" % str([t.status for t in self]))
        #       self.Error(msg)

        return self.work.returncodes

##########################################################################################



#!/usr/bin/env python

"""
Module which defines basic entries for each ion and oxide to compute a
Pourbaix diagram
"""

from __future__ import division

__author__ = "Sai Jayaraman"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.0"
__maintainer__ = "Sai Jayaraman"
__email__ = "sjayaram@mit.edu"
__status__ = "Development"
__date__ = "December 10, 2012"

import re
import math

from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Composition
from pymatgen.serializers.json_coders import MSONable
from pymatgen.core.ion import Ion
from pymatgen.phasediagram.entries import PDEntry

PREFAC = 0.0591


class PourbaixEntry(MSONable):
    """
    An object encompassing all data relevant to an ion in a pourbaix diagram.
    Each bulk solid/ion has a free energy g of the form:
    g = g0_ref + 0.0591 log10(conc) - nO mu_H2O + (nH - 2nO) pH
    + phi (-nH + 2nO + q)
    """
    def __init__(self, entry, correction=0.0, entry_id=None):
        """
        Args:
            entry:
                An entry object
                (ComputedEntry/ComputedStructureEntry/PDEntry/IonEntry)
            energy:
                Energy of entry
        """
        if isinstance(entry, IonEntry):
            self._entry = entry
            self._conc = 1.0e-6
            self._phase_type = "Ion"
            self._charge = entry.composition.charge
        else:
            self._entry = entry
            self._conc = 1.0
            self._phase_type = "Solid"
            self._charge = 0.0
        self._npH = None
        self._nPhi = None
        self._nH2O = None
        self._nM = None
        self.uncorrected_energy = entry.energy
        self.correction = correction
        self._calc_coeff_terms()
        self._name = self._entry.composition.reduced_formula
        if self._phase_type == "Solid":
            self._name += "(s)"
        try:
            self.entry_id = entry.entry_id
        except AttributeError:
            self.entry_id = entry_id

    @property
    def _g0(self):
        return self.energy

    @property
    def energy(self):
        return self.uncorrected_energy + self.correction

    @property
    def name(self):
        """
        Returns the entry's name
        """
        return self._name

    def set_name(self, string):
        """
        Set name of entry

        Args:
            string: Input string
        """
        self._name = string

    @property
    def npH(self):
        """
        Returns value of npH, the coefficient of pH
        """
        return self._npH

    @property
    def nH2O(self):
        """
        Returns coefficient of Mu_H2O
        """
        return self._nH2O

    @property
    def nPhi(self):
        """
        Returns nPhi, the coefficient of Phi
        """
        return self._nPhi

    @property
    def g0(self):
        """
        Return g0 for the entry. Legacy function.
        """
        return self._g0

    @property
    def conc(self):
        """
        Return concentration of the entry. Returns 1 if solid.
        """
        return self._conc
    
    @property
    def conc_term(self):
        """
        Returns the concentration contribution to the free energy.
        """
        return self.normalization_factor * PREFAC * math.log10(self._conc)

    @property
    def phase_type(self):
        """
        Returns whether the entry is a solid/ion.
        """
        return self._phase_type

    def g0_add(self, term):
        """
        Add a correction term to g0.

        Args:
            term:
                Correction term to add to g0
        """
        self.correction += term

    def g0_replace(self, term):
        """
        Replace g0 by a different value.

        Args:
            term: 
                New value for g0
        """
        self.uncorrected_energy = term
        self.correction = 0.0

    @property
    def to_dict(self):
        """
        Returns dict which contains Pourbaix Entry data.
        Note that the pH, voltage, H2O factors are always calculated when
        constructing a PourbaixEntry object.
        """
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__}
        if isinstance(self._entry, IonEntry):
            d["entry type"] = "Ion"
        else:
            d["entry type"] = "Solid"
        d["entry"] = self._entry.to_dict
        d["pH factor"] = self._npH
        d["voltage factor"] = self._nPhi
        d["concentration"] = self._conc
        d["H2O factor"] = self._nH2O
        d["energy"] = self.energy
        d["correction"] = self.correction
        d["entry_id"] = self.entry_id
        return d

    @classmethod
    def from_dict(cls, d):
        """
        Returns a PourbaixEntry by reading in an Ion
        """
        entry_type = d["entry type"]
        if entry_type == "Ion":
            entry = IonEntry.from_dict(d["entry"])
        else:
            entry = PDEntry.from_dict(d["entry"])
        correction = d["correction"]
        entry_id = d["entry_id"]
        return PourbaixEntry(entry, correction, entry_id)

    def _calc_coeff_terms(self):
        """
        Calculates coefficients of pH, V, H2O
        """
        nH = 0
        nO = 0
        nM = 0
        for elt in self._entry.composition.elements:
            if elt == (Element("H")):
                nH = self.entry.composition[elt]
            elif elt == (Element("O")):
                nO = self.entry.composition[elt]
            else:
                nM += self.entry.composition[elt]
        self._nM = nM
        self._npH = (nH - 2 * nO)
        self._nH2O = nO
        self._nPhi = (nH - 2 * nO - self._charge)

    @property
    def normalization_factor(self):
        """
        Normalize each entry by nM
        """
        fact = 1.0 / self._nM
        return fact

    def scale(self, factor):
        """
        Normalize all entries by normalization factor.

        Args:
            factor:
                Normalization factor
        """
        self._npH *= factor
        self._nPhi *= factor
        self._nH2O *= factor
        self.uncorrected_energy *= factor
        self.correction *= factor
#        self._g0 *= factor

    def normalize(self, factor):
        self.scale(factor)

    @property
    def charge(self):
        """
        Returns charge of entry
        """
        return self._charge

    @property
    def composition(self):
        """
        Returns composition
        """
        return self.entry.composition

    @property
    def entry(self):
        """
        Returns IonEntry/PDEntry object
        """
        return self._entry

    def reduced_entry(self):
        """
        Calculate reduction factor for composition, and reduce parameters by
        this factor.
        """
        reduction_factor = self.entry.composition.\
            get_reduced_composition_and_factor()[1]
        self._nM /= reduction_factor
        self.scale(1.0 / reduction_factor)

    @property
    def num_atoms(self):
        """
        Return number of atoms in current formula. Useful for normalization
        """
        return self.entry.composition.num_atoms\
            / self.entry.composition.get_reduced_composition_and_factor()[1]

    def set_conc(self, conc):
        """
        Set concentration manually.

        Args:
            conc:
                Input concentration
        """
        self._conc = conc

    def __repr__(self):
        return "Pourbaix Entry : {} with energy = {:.4f}, npH = {}, nPhi = {},\
             nH2O = {}".format(self._entry.composition, self.g0, self.npH,
                               self.nPhi, self.nH2O)

    def __str__(self):
        return self.__repr__()


class MultiEntry(PourbaixEntry):
    """
    PourbaixEntry-like object for constructing multi-elemental Pourbaix
    diagrams.
    """
    def __init__(self, entry_list, weights=None):
        """
        Args:
            entry_list:
                List of component PourbaixEntries
            weights:
                Weights associated with each entry. Default is None
        """
        if weights is None:
            self._weights = [1.0] * len(entry_list)
        else:
            self._weights = weights
        self._entrylist = entry_list
        self.correction = 0.0
        self.uncorrected_energy = 0.0
        self._npH = 0.0
        self._nPhi = 0.0
        self._nH2O = 0.0
        self._nM = 0.0
        self._name = ""
        self.entry_id = list()
        for i in xrange(len(entry_list)):
            entry = entry_list[i]
            self.uncorrected_energy += self._weights[i] * \
                entry.uncorrected_energy
            self.correction += self._weights[i] * entry.correction
            self._npH += self._weights[i] * entry.npH
            self._nPhi += self._weights[i] * entry.nPhi
            self._nH2O += self._weights[i] * entry.nH2O
            self._nM += self._weights[i] * entry._nM
            self._name += entry.name + " + "
            self.entry_id.append(entry.entry_id)
        self._name = self._name[:-3]

    @property
    def normalization_factor(self):
        """
        Normalize each entry by nM
        """
        norm_fac = 0.0
        for i in xrange(len(self._entrylist)):
            entry = self._entrylist[i]
            for el in entry.composition.elements:
                if (el == Element("O")) | (el == Element("H")):
                    continue
                if entry._phase_type == 'Solid':
                    red_fac = entry.composition.\
                        get_reduced_composition_and_factor()[1]
                else:
                    red_fac = 1.0
                norm_fac += self._weights[i] * entry.composition[el] / red_fac
        fact = 1.0 / norm_fac
        return fact

    def __repr__(self):
        str = "Multiple Pourbaix Entry : with energy = {:.4f}, npH = {}, "\
            "nPhi = {}, nH2O = {}".format(
            self.g0, self.npH, self.nPhi, self.nH2O)
        str += ", species: "
        for entry in self._entrylist:
            str += entry.name + " + "
        return str[:-3]

    def __str__(self):
        return self.__repr__()

    @property
    def conc_term(self):
        sum_conc = 0.0
        for i in xrange(len(self._entrylist)):
            entry = self._entrylist[i]
            sum_conc += self._weights[i] * PREFAC * math.log10(entry.conc)
        return sum_conc * self.normalization_factor

    @property
    def entrylist(self):
        return self._entrylist


class IonEntry(PDEntry):
    """
    Object similar to PDEntry, but contains an Ion object instead of a
    Composition object.

    .. attribute:: name

        A name for the entry. This is the string shown in the phase diagrams.
        By default, this is the reduced formula for the composition, but can be
        set to some other string for display purposes.
    """
    def __init__(self, ion, energy, name=None):
        """
        Args:
            comp:
                Ion object
            energy:
                Energy for composition.
            name:
                Optional parameter to name the entry. Defaults to the
                chemical formula.
        """
        self._energy = energy
        self._composition = ion
        self.name = name if name else self._composition.reduced_formula

    @classmethod
    def from_dict(cls, d):
        """
        Returns an IonEntry object from a dict.
        """
        return IonEntry(Ion.from_dict(d["composition"]), d["energy"])

    @property
    def to_dict(self):
        """
        Creates a dict of composition, energy, and ion name
        """
        d = {"composition": self._composition.to_dict, "energy": self._energy}
        return d

    @property
    def energy(self):
        """
        Return final energy
        """
        return self._energy

    @property
    def energy_per_atom(self):
        """
        Return final energy per atom
        """
        return self._energy / self.composition.num_atoms

    @property
    def composition(self):
        """
        Returns the composition
        """
        return self._composition

    def __repr__(self):
        return "IonEntry : {} with energy = {:.4f}".format(self.composition,
                                                           self.energy)

    def __str__(self):
        return self.__repr__()


class PourbaixEntryIO(object):
    """
    Class to import and export Pourbaix entries from a csv file
    """
    @staticmethod
    def to_csv(filename, entries, latexify_names=False):
        """
        Exports Pourbaix entries to a csv

        Args:
            filename:
                Filename to write to.
            entries:
                Entries to export.
            latexify_names:
                Format entry names to be LaTex compatible, e.g., Li_{2}O
        """
        import csv
        elements = set()
        map(elements.update, [entry.entry.composition.elements
                              for entry in entries])
        elements = sorted(list(elements), key=lambda a: a.X)
        writer = csv.writer(open(filename, "wb"), delimiter=",",
                            quotechar="\"", quoting=csv.QUOTE_MINIMAL)
        writer.writerow(["Name"] + elements + ["Energy"] + ["Entry Type"]
                        + ["Charge"] + ["Concentration"])
        for entry in entries:
            row = [entry.name if not latexify_names
                   else re.sub(r"([0-9]+)", r"_{\1}", entry.name)]
            if entry.phase_type == "Solid":
                reduction_fac = entry.entry.composition.\
                    get_reduced_composition_and_factor()[1]
            else:
                reduction_fac = 1.0
            row.extend([entry.entry.composition[el] / reduction_fac
                        for el in elements])
            if entry.phase_type == "Solid":
                reduction_fac = 1.0
            row.append(entry.g0 / reduction_fac)
            row.append(entry.phase_type)
            row.append(entry.charge / reduction_fac)
            row.append(entry.conc)
            writer.writerow(row)

    @staticmethod
    def from_csv(filename):
        """
        Imports PourbaixEntries from a csv.

        Args:
            filename - Filename to import from.

        Returns:
            List of Entries
        """
        import csv
        reader = csv.reader(open(filename, "rb"), delimiter=",",
                            quotechar="\"", quoting=csv.QUOTE_MINIMAL)
        entries = list()
        header_read = False
        for row in reader:
            if not header_read:
                elements = row[1:(len(row) - 4)]
                header_read = True
            else:
                name = row[0]
                energy = float(row[-4])
                conc = float(row[-1])
                comp = dict()
                for ind in range(1, len(row) - 4):
                    if float(row[ind]) > 0:
                        comp[Element(elements[ind - 1])] = float(row[ind])
                phase_type = row[-3]
                if phase_type == "Ion":
                    PoE = PourbaixEntry(IonEntry(Ion.from_formula(name),
                                                 energy))
                    PoE.set_conc(conc)
                    PoE.set_name(name)
                    entries.append(PoE)
                else:
                    entries.append(PourbaixEntry(PDEntry(Composition(comp),
                                                         energy)))
        elements = [Element(el) for el in elements]
        return elements, entries



#!/usr/bin/env python

"""
This module contains useful decorators for a variety of functions.
"""

from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Dec 31, 2011"

import logging
import datetime
from functools import wraps
import warnings


def singleton(cls):
    """
    This decorator can be used to create a singleton out of a class.
    """

    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance


def cached_class(klass):
    """
    Decorator to cache class instances by constructor arguments.
    This results in a class that behaves like a singleton for each
    set of constructor arguments, ensuring efficiency.

    Note that this should be used for *immutable classes only*.  Having
    a cached mutable class makes very little sense.  For efficiency,
    avoid using this decorator for situations where there are many
    constructor arguments permutations.

    The keywords argument dictionary is converted to a tuple because
    dicts are mutable; keywords themselves are strings and
    so are always hashable, but if any arguments (keyword
    or positional) are non-hashable, that set of arguments
    is not cached.
    """
    cache = {}

    @wraps(klass, assigned=("__name__", "__module__"), updated=())
    class _decorated(klass):
        # The wraps decorator can't do this because __doc__
        # isn't writable once the class is created
        __doc__ = klass.__doc__

        def __new__(cls, *args, **kwds):
            #key = (cls,) + args + tuple(kwds.items())
            #if key not in cache:
            #    o = super(klass, cls).__new__(cls, *args, **kwds)
            #    cache[key] = o
            #return cache[key]
            key = (cls,) + args + tuple(kwds.iteritems())
            try:
                inst = cache.get(key, None)
            except TypeError:
                # Can't cache this set of arguments
                inst = key = None
            if inst is None:
                # Technically this is cheating, but it works,
                # and takes care of initializing the instance
                # (so we can override __init__ below safely);
                # calling up to klass.__new__ would be the
                # "official" way to create the instance, but
                # that raises DeprecationWarning if there are
                # args or kwds and klass does not override
                # __new__ (which most classes don't), because
                # object.__new__ takes no parameters (and in
                # Python 3 the warning will become an error)
                inst = klass(*args, **kwds)
                # This makes isinstance and issubclass work
                # properly
                inst.__class__ = cls
                if key is not None:
                    cache[key] = inst
            return inst

        def __init__(self, *args, **kwds):
            # This will be called every time __new__ is
            # called, so we skip initializing here and do
            # it only when the instance is created above
            pass

    return _decorated


def logged(level=logging.DEBUG):
    """
    Useful logging decorator. If a method is logged, the beginning and end of
    the method call will be logged at a pre-specified level.

    Args:
        level:
            Level to log method at. Defaults to DEBUG.
    """
    def wrap(f):
        logger = logging.getLogger("{}.{}".format(f.__module__, f.__name__))

        def wrapped_f(*args, **kwargs):
            logger.log(level, "Called at {} with args = {} and kwargs = {}"
                       .format(datetime.datetime.now(), args, kwargs))
            data = f(*args, **kwargs)
            logger.log(level, "Done at {} with args = {} and kwargs = {}"
                       .format(datetime.datetime.now(), args, kwargs))
            return data

        return wrapped_f
    return wrap


def deprecated(replacement=None):
    """
    Decorator to mark classes or functions as deprecated,
    with a possible replacement.

    Args:
        replacement:
            A replacement class or method.

    Returns:
        Original function, but with a warning to use the updated class.
    """
    def wrap(old):
        def wrapped(*args, **kwargs):
            msg = "{} is deprecated".format(old.__name__)
            if replacement is not None:
                msg += "; use {} in {} instead.".format(
                    replacement.__name__, replacement.__module__)
            warnings.simplefilter('default')
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return old(*args, **kwargs)
        return wrapped
    return wrap


class requires(object):
    """
    Decorator to mark classes or functions as requiring a specified condition
    to be true. This can be used to present useful error messages for
    optional dependencies. For example, decorating the following code will
    check if scipy is present and if not, a runtime error will be raised if
    someone attempts to call the use_scipy function::

        try:
            import scipy
        except ImportError:
            scipy = None

        @requires(scipy is not None, "scipy is not present.")
        def use_scipy():
            print scipy.majver

    Args:
        condition:
            Condition necessary to use the class or function.
        message:
            A message to be displayed if the condition is not True.
    """

    def __init__(self, condition, message):
        self.condition = condition
        self.message = message

    def __call__(self, callable):
        @wraps(callable)
        def decorated(*args, **kwargs):
            if not self.condition:
                raise RuntimeError(self.message)
            return callable(*args, **kwargs)
        return decorated



"""
This module provides utilities for basic math operations.
"""
from __future__ import division, print_function

import itertools
import collections
import numpy as np


def iterator_from_slice(s):
    """
    Constructs an iterator given a slice object s.

    .. note::

        The function returns an infinite iterator if s.stop is None
    """
    start = s.start if s.start is not None else 0
    step = s.step if s.step is not None else 1

    if s.stop is None:
        # Infinite iterator.
        return itertools.count(start=start, step=step)
    else:
        # xrange-like iterator that suppors float.
        return iter(np.arange(start, s.stop, step))


def sort_dict(d, key=None, reverse=False):
    """
    Sorts a dict by value.

    Args:
        d:
            Input dictionary
        key:
            function which takes an tuple (key, object) and returns a value to
            compare and sort by. By default, the function compares the values
            of the dict i.e. key = lambda t : t[1]
        reverse:
            allows to reverse sort order.

    Returns:
        OrderedDict object whose keys are ordered according to their value.
    """
    kv_items = [kv for kv in d.items()]

    # Sort kv_items according to key.
    if key is None:
        kv_items.sort(key=lambda t: t[1], reverse=reverse)
    else:
        kv_items.sort(key=key, reverse=reverse)

    # Build ordered dict.
    return collections.OrderedDict(kv_items)


def chunks(items, n):
    """
    Yield successive n-sized chunks from a list-like object.

    >>> import pprint
    >>> pprint.pprint(list(chunks(range(1, 25), 10)))
    [[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
     [11, 12, 13, 14, 15, 16, 17, 18, 19, 20],
     [21, 22, 23, 24]]
    """
    for i in range(0, len(items), n):
        yield items[i:i+n]


def min_max_indexes(seq):
    """
    Uses enumerate, max, and min to return the indices of the values
    in a list with the maximum and minimum value:
    """
    minimum = min(enumerate(seq), key=lambda s: s[1])
    maximum = max(enumerate(seq), key=lambda s: s[1])
    return minimum[0], maximum[0]



__author__ = ", ".join(["Shyue Ping Ong", "Anubhav Jain", "Geoffroy Hautier",
                        "William Davidson Richard", "Stephen Dacek",
                        "Michael Kocher", "Dan Gunter", "Shreyas Cholia",
                        "Vincent L Chevrier", "Rickard Armiento"])
__date__ = "Sep 6 2013"
__version__ = "2.8.0"

#Useful aliases for commonly used objects and modules.

from .core import *
from .serializers.json_coders import PMGJSONEncoder, PMGJSONDecoder, \
    pmg_dump, pmg_load
from .electronic_structure.core import Spin, Orbital
from .util.io_utils import zopen
from .io.smartio import read_structure, write_structure, read_mol, write_mol
from .matproj.rest import MPRester



from __future__ import division, print_function

import os
import cStringIO as StringIO

from subprocess import Popen, PIPE

from pymatgen.util.io_utils import which
from pymatgen.util.string_utils import list_strings

__author__ = "Matteo Giantomassi"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"
__email__ = "gmatteo at gmail.com"
__status__ = "Development"
__date__ = "$Feb 21, 2013M$"

__all__ = [
    "Mrgscr",
    "Mrggkk",
    "Mrgdbb",
    "Anaddb",
]


class ExecWrapper(object):
    """This class runs an executable in a subprocess."""
    def __init__(self, executable, verbose=0):
        """
        Args:
            executable:
                path to the executable.
            verbose:
                Verbosity level.
        """
        if executable is None: executable = self.name

        self.executable = which(executable)

        self.verbose = int(verbose)

        if self.executable is None:
            msg = "Cannot find executable %s is PATH\n Use export PATH=/dir_with_exec:$PATH" % executable
            raise self.Error(msg)

        assert os.path.basename(self.executable) == self._name

    def __str__(self):
        return "%s" % self.executable

    @property
    def name(self):
        return self._name

    def execute(self, cwd=None, **kwargs):
        """Execute the executable in a subprocess."""
        args = [self.executable, "<", self.stdin_fname, ">", self.stdout_fname, "2>", self.stderr_fname]

        self.cmd_str = " ".join(args)

        p = Popen(self.cmd_str, shell=True, stdout=PIPE, stderr=PIPE, cwd=cwd)

        (self.stdout_data, self.stderr_data) = p.communicate()

        self.returncode = p.returncode

        if self.returncode != 0:
            with open(self.stdout_fname, "r") as out, open(self.stderr_fname, "r") as err:
                self.stdout_data = out.read()
                self.stderr_data = err.read()

            if self.verbose:
                print("*** stdout: ***\n", self.stdout_data)
                print("*** stderr  ***\n", self.stderr_data)

            raise self.Error("%s returned %s\n cmd_str: %s" % (self, self.returncode, self.cmd_str))

##########################################################################################

class MrgscrError(Exception):
    """Error class for Mrgscr"""

class Mrgscr(ExecWrapper):
    _name = "mrgscr"

    Error = MrgscrError

    def merge_qpoints(self, files_to_merge, out_prefix, cwd=None):
        """
        Execute mrgscr in a subprocess to merge files_to_merge. Produce new file with prefix out_prefix

        If cwd is not None, the child's current directory will be changed to cwd before it is executed.
        """
        # We work with absolute paths.
        files_to_merge = [os.path.abspath(s) for s in list_strings(files_to_merge)]
        nfiles = len(files_to_merge)

        if self.verbose:
            print("Will merge %d files with output_prefix %s" % (nfiles, out_prefix))
            for (i, f) in enumerate(files_to_merge):
                print(" [%d] %s" % (i, f))

        if nfiles == 1:
            raise self.Error("merge_qpoints does not support nfiles == 1")

        self.stdin_fname, self.stdout_fname, self.stderr_fname = (
            "mrgscr.stdin", "mrgscr.stdout", "mrgscr.stderr")
        if cwd is not None:
            self.stdin_fname, self.stdout_fname, self.stderr_fname = \
                map(pj, 3 * [cwd], [self.stdin_fname, self.stdout_fname, self.stderr_fname])

        inp = StringIO.StringIO()

        inp.write(str(nfiles) + "\n")      # Number of files to merge.
        inp.write(out_prefix + "\n")      # Prefix for the final output file:

        for filename in files_to_merge:
            inp.write(filename + "\n")   # List with the files to merge.

        inp.write("1\n")                 # Option for merging q-points.

        inp.seek(0)
        self.stdin_data = [s for s in inp]

        with open(self.stdin_fname, "w") as fh:
            fh.writelines(self.stdin_data)

        try:
            self.execute(cwd=cwd)
        except self.Error:
            raise

##########################################################################################

class MrggkkError(Exception):
    """Error class for Mrggkk."""


class Mrggkk(ExecWrapper):
    _name = "mrggkk"

    Error = MrggkkError

    def merge(self, gswfk_file, dfpt_files, gkk_files, out_fname, binascii=0, cwd=None):
        """Merge DDB file, return the absolute path of the new database."""
        raise NotImplementedError("This method should be tested")

        out_fname = out_fname if cwd is None else os.path.join(os.path.abspath(cwd), out_fname)

        # We work with absolute paths.
        gswfk_file = absath(gswfk_file)
        dfpt_files = [os.path.abspath(s) for s in list_strings(dfpt_files)]
        gkk_files = [os.path.abspath(s) for s in list_strings(gkk_files)]

        if self.verbose:
            print("Will merge %d 1WF files, %d GKK file in output %s" %
                  (len(dfpt_nfiles), (len_gkk_files), out_fname))
            for (i, f) in enumerate(dfpt_files): print(" [%d] 1WF %s" % (i, f))
            for (i, f) in enumerate(gkk_files):  print(" [%d] GKK %s" % (i, f))

        self.stdin_fname, self.stdout_fname, self.stderr_fname = (
            "mrggkk.stdin", "mrggkk.stdout", "mrggkk.stderr")
        if cwd is not None:
            self.stdin_fname, self.stdout_fname, self.stderr_fname = \
                map(pj, 3 * [cwd], [self.stdin_fname, self.stdout_fname, self.stderr_fname])

        inp = StringIO.StringIO()

        inp.write(out_fname + "\n")        # Name of the output file
        inp.write(str(binascii) + "\n")    # Integer flag: 0 --> binary output, 1 --> ascii formatted output
        inp.write(gswfk_file + "\n")       # Name of the groud state wavefunction file WF

        #dims = len(dfpt_files, gkk_files, ?)
        dims = " ".join([str(d) for d in dims])
        inp.write(dims + "\n")             # Number of 1WF, of GKK files, and number of 1WF files in all the GKK files

        # Names of the 1WF files...
        for fname in dfpt_files: inp.write(fname + "\n")

        # Names of the GKK files...
        for fname in gkk_files: inp.write(fname + "\n")

        inp.seek(0)
        self.stdin_data = [s for s in inp]

        with open(self.stdin_fname, "w") as fh:
            fh.writelines(self.stdin_data)

        try:
            self.execute(cwd=cwd)
        except self.Error:
            raise

        return out_fname

##########################################################################################

class MrgddbError(Exception):
    """Error class for Mrgddb."""

class Mrgddb(ExecWrapper):
    _name = "mrgddb"

    Error = MrgddbError

    def merge(self, ddb_files, out_ddb, description, cwd=None):
        """Merge DDB file, return the absolute path of the new database."""

        # We work with absolute paths.
        ddb_files = [os.path.abspath(s) for s in list_strings(ddb_files)]

        out_ddb = out_ddb if cwd is None else os.path.join(os.path.abspath(cwd), out_ddb)

        if self.verbose:
            print("Will merge %d files into output DDB %s" % (len(ddb_files), out_ddb))
            for (i, f) in enumerate(ddb_files):
                print(" [%d] %s" % (i, f))

        # Handle the case of a single file since mrgddb uses 1 to denote GS files!
        if len(ddb_files) == 1:
            with open(ddb_files[0], "r") as inh, open(out_ddb, "w") as out:
                for line in inh:
                    out.write(line)
            return out_ddb

        self.stdin_fname, self.stdout_fname, self.stderr_fname = (
            "mrgddb.stdin", "mrgddb.stdout", "mrgddb.stderr")

        if cwd is not None:
            self.stdin_fname, self.stdout_fname, self.stderr_fname = \
                map(pj, 3 * [cwd], [self.stdin_fname, self.stdout_fname, self.stderr_fname])

        inp = StringIO.StringIO()

        inp.write(out_ddb + "\n")            # Name of the output file.
        inp.write(str(description) + "\n")     # Description.
        inp.write(str(len(ddb_files)) + "\n")  # Number of input DDBs.

        # Names of the DDB files.
        for fname in ddb_files:
            inp.write(fname + "\n")

        inp.seek(0)
        self.stdin_data = [s for s in inp]

        with open(self.stdin_fname, "w") as fh:
            fh.writelines(self.stdin_data)

        try:
            self.execute(cwd=cwd)
        except self.Error:
            raise

        return out_ddb

##########################################################################################

class AnaddbError(Exception):
    """Error class for Anaddb."""

class Anaddb(ExecWrapper):
    _name = "anaddb"

    Error = AnaddbError

    #def make_stdin(self):
    #    # Files file
    #    inp = StringIO.StringIO()
    #    inp.write(self.input_fname + "\n")     # Input file.
    #    inp.write(self.stdout_fname + "\n")    # Output file.
    #    inp.write(ddb_file + "\n")             # DDB file
    #    inp.write("dummy_band2eps" + "\n")
    #    inp.write("dummy1" + "\n")
    #    inp.write("dummy2" + "\n")
    #    inp.write("dummy3" + "\n")
    #    inp.seek(0)
    #    self.stdin_data = [s for s in inp]

    def diagonalize_1q(self, ddb_file, cwd=None):

        # We work with absolute paths.
        ddb_file = os.path.abspath(ddb_file)

        self.stdin_fname, self.input_fname, self.stdout_fname, self.stderr_fname = (
            "anaddb.stdin", "anaddb.input", "anaddb.stdout", "anaddb.stderr")

        if cwd is not None:
            self.stdin_fname, self.input_fname, self.stdout_fname, self.stderr_fname = \
                map(pj, 3 * [cwd], [self.stdin_fname, self.inp_fname, self.stdout_fname, self.stderr_fname])

        # Files file
        inp = StringIO.StringIO()

        inp.write(self.input_fname + "\n")     # Input file.
        inp.write(self.stdout_fname + "\n")    # Output file.
        inp.write(ddb_file + "\n")             # DDB file
        inp.write("dummy_band2eps" + "\n")
        inp.write("dummy1" + "\n")
        inp.write("dummy2" + "\n")
        inp.write("dummy3" + "\n")

        inp.seek(0)
        self.stdin_data = [s for s in inp]

        with open(self.stdin_fname, "w") as fh:
            fh.writelines(self.stdin_data)

        # Get the q-point from the DDB file
        with open(ddb_file, "r") as fh:
            nfound = 0
            tag = " qpt "
            for line in fh:
                print(line)
                if line.startswith(tag):
                    nfound += 1
                    # Coordinates of the q-points.
                    qcoords_str = line.split()[1:4]
                    #qcoords_str = [ s.replace("D", "E") for s in qcoords_str]
                    qpoint = map(float, qcoords_str)

        if nfound != 1:
            raise self.Error("Found %s occurrences of tag %s in file %s" % (nfound, tag, ddb_file))

        # Write simple input file for the anaddb code.
        with open(self.input_fname, "w") as inp:
            inp.write('# Flags\n')
            inp.write(' ifcflag   1 # Interatomic force constant flag\n\n')
            inp.write('# Wavevector grid number 1 (coarse grid, from DDB)\n\n')
            inp.write(' brav    1         # Bravais Lattice : 1-S.C., 2-F.C., 3-B.C., 4-Hex.\n')
            inp.write(' ngqpt   1  1  1   # Q-mesh\n')
            inp.write(' nqshft  1         # number of q-shifts\n')
            inp.write(' q1shft  %f %f %f' % tuple(qpoint))

            #inp.write('# Effective charges
            #inp.write('     asr   1     ! Acoustic Sum Rule. 1 => imposed asymetrically
            #inp.write('  chneut   1     ! Charge neutrality requirement for effective charges.
            #inp.write('# Interatomic force constant info
            #inp.write('  dipdip  1      ! Dipole-dipole interaction treatment
            #inp.write('  ifcana  1      ! Analysis of the IFCs
            #inp.write('  ifcout 20      ! Number of IFC's written in the output, per atom
            #inp.write('  natifc  1      ! Number of atoms in the cell for which ifc's are analysed
            #inp.write('   atifc  1      ! List of atoms
            #inp.write('
            #inp.write('# This line added when defaults were changed (v5.3) to keep the previous, old behaviour
            #inp.write('#  symdynmat 0

        if self.verbose:
            print("Will diagonalize DDB file : %s" % ddb_file)

        try:
            self.execute(cwd=cwd)
        except self.Error:
            raise

        # Get frequencies from the output file
        # TODO
        #with open(self.stdout_fname, "r") as out:
            #print(out.readlines())
            #for line in out:
            #    if line: raise
            #return frequencies



"""
Common test support for pymatgen test scripts.

This single module should provide all the common functionality for pymatgen
tests in a single location, so that test scripts can just import it and work
right away.
"""
import unittest
import numpy.testing.utils as nptu


class PymatgenTest(unittest.TestCase):
    """
    Extends unittest.TestCase with functions (taken from numpy.testing.utils)
    that support the comparison of arrays.
    """

    @staticmethod
    def assert_almost_equal(actual, desired, decimal=7, err_msg='',
                            verbose=True):
        """
        Alternative naming for assertArrayAlmostEqual.
        """
        return PymatgenTest.assertArrayAlmostEqual(
            actual, desired, decimal, err_msg, verbose)

    @staticmethod
    def assert_equal(actual, desired, err_msg='', verbose=True):
        """
        Alternative naming for assertArrayEqual.
        """
        return PymatgenTest.assertArrayEqual(actual, desired,
                                             err_msg=err_msg, verbose=verbose)

    @staticmethod
    def assertArrayAlmostEqual(actual, desired, decimal=7, err_msg='',
                               verbose=True):
        """
        Tests if two arrays are almost equal to a tolerance. The CamelCase
        naming is so that it is consistent with standard unittest methods.
        """
        return nptu.assert_almost_equal(actual, desired, decimal, err_msg,
                                        verbose)

    @staticmethod
    def assertArrayEqual(actual, desired, err_msg='', verbose=True):
        """
        Tests if two arrays are equal. The CamelCase naming is so that it is
         consistent with standard unittest methods.
        """
        return nptu.assert_equal(actual, desired, err_msg=err_msg,
                                 verbose=verbose)



"""
Classes defining Abinit calculations and workflows
"""
from __future__ import division, print_function

import os
import shutil
import collections
import abc
import numpy as np

from pymatgen.core.design_patterns import Enum, AttrDict
from pymatgen.util.string_utils import stream_has_colours
from pymatgen.serializers.json_coders import MSONable, json_load, json_pretty_dump

from pymatgen.io.abinitio.utils import abinit_output_iscomplete, File
from pymatgen.io.abinitio.launcher import TaskLauncher
from pymatgen.io.abinitio.events import EventParser

#import logging
#logger = logging.getLogger(__name__)

__author__ = "Matteo Giantomassi"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"

__all__ = [
    "task_factory",
]


class FakeProcess(object):
    """
    This object is attached to a Task instance if the task has not been submitted
    This trick allows us to simulate a process that is still running so that we can safely poll task.process.
    """
    def poll(self):
        return None

    def wait(self):
        raise RuntimeError("Cannot wait a FakeProcess")

    def communicate(self, input=None):
        raise RuntimeError("Cannot communicate with a FakeProcess")

    def kill(self):
        raise RuntimeError("Cannot kill a FakeProcess")

    @property
    def returncode(self):
        return None


def task_factory(strategy, workdir, runmode, task_id=1, links=None, **kwargs):
    """Factory function for Task instances."""
    # TODO
    # Instanciate subclasses depending on the runlevel.
    classes = {
       "scf": AbinitTask,
       "nscf": AbinitTask,
       "relax": AbinitTask,
       "dfpt": AbinitTask,
       "screening": AbinitTask,
       "sigma": AbinitTask,
       "bse": AbinitTask,
    }

    return classes[strategy.runlevel](strategy, workdir, runmode, task_id=task_id, links=links, **kwargs)


class TaskError(Exception):
    """Base Exception for Task methods"""


class Task(object):
    __metaclass__ = abc.ABCMeta

    Error = TaskError

    # Possible status of the task.
    S_DONE = 1   # Task completed, This does not imply that results are ok or that the calculation completed successfully
    S_READY = 2  # Task is ready for submission.
    S_SUB = 4    # Task has been submitted.
    S_RUN = 8    # Task is running.

    status2str = {
        S_DONE: "done",
        S_READY: "ready",
        S_SUB: "submitted",
        S_RUN: "running",
    }

    def __init__(self):
        # Set the initial status.
        self.set_status(Task.S_READY)

    # Interface modeled after subprocess.Popen
    @abc.abstractproperty
    def process(self):
        """Return an object that supports the `subprocess.Popen` protocol."""

    def poll(self):
        """Check if child process has terminated. Set and return returncode attribute."""
        returncode = self.process.poll()
        if returncode is not None:
            self.set_status(Task.S_DONE)

        return returncode

    def wait(self):
        """Wait for child process to terminate. Set and return returncode attribute."""
        returncode = self.process.wait()
        self.set_status(Task.S_DONE)

        return returncode

    def communicate(self, input=None):
        """
        Interact with process: Send data to stdin. Read data from stdout and stderr, until end-of-file is reached. 
        Wait for process to terminate. The optional input argument should be a string to be sent to the 
        child process, or None, if no data should be sent to the child.

        communicate() returns a tuple (stdoutdata, stderrdata).
        """
        stdoutdata, stderrdata = self.process.communicate(input=input)
        self.set_status(Task.S_DONE)
        return stdoutdata, stderrdata 

    def kill(self):
        """Kill the child."""
        self.process.kill()

    @property
    def returncode(self):
        """
        The child return code, set by poll() and wait() (and indirectly by communicate()). 
        A None value indicates that the process hasn't terminated yet.
        A negative value -N indicates that the child was terminated by signal N (Unix only).
        """
        return self.process.returncode

    @property
    def id(self):
        """Task identifier."""
        return self._id

    def set_id(self, id):
        """Set the task identifier."""
        self._id = id

    @property
    def tot_ncpus(self):
        """Total number of CPUs used to run the task."""
        return self.launcher.tot_ncpus

    @property
    def status(self):
        """Gives the status of the task."""
        return self._status

    def set_status(self, status):
        """Set the status of the task."""
        if status not in Task.status2str:
            raise RuntimeError("Unknown status: %s" % status)

        self._status = status

        # Notify the observers
        #self.subject.notify_observers(status)

    @property
    def links_status(self):
        """Returns a list with the status of the links."""
        if not self._links:
            return [Task.S_DONE,]
        return [l.status for l in self._links]

    @abc.abstractmethod
    def setup(self, *args, **kwargs):
        """Public method called before submitting the task."""

    def _setup(self, *args, **kwargs):
        """
        This method calls self.setup after having performed additional operations
        such as the creation of the symbolic links needed to connect different tasks.
        """
        # Create symbolic links to the output files produced by the other tasks
        for link in self._links:
            filepaths, exts = link.get_filepaths_and_exts()
            for (path, ext) in zip(filepaths, exts):
                dst = self.idata_path_from_ext(ext)

                if not os.path.exists(path): 
                    # Try netcdf file.
                    # TODO: this case should be treated in a cleaner way.
                    path += "-etsf.nc"
                    if os.path.exists(path):
                        dst += "-etsf.nc"

                if not os.path.exists(path):
                    raise self.Error("%s is needed by this task but it does not exist" % path)

                # Link path to dst
                #print("Linking ", path," --> ",dst)
                os.symlink(path, dst)

        self.setup(*args, **kwargs)

    @property
    def events(self):
        """List of errors or warnings reported by ABINIT."""
        if self.status != Task.S_DONE:
            raise self.Error(
                "Task %s is not completed.\n You cannot access its events now, use get_events" % repr(self))
        try:
            return self._events
        except AttributeError:
            self._events = self.get_events()
            return self._events

    def get_events(self):
        """Analyzes the main output for possible errors or warnings."""
        # TODO: Handle possible errors in the parser by generating a custom EventList object
        return EventParser().parse(self.output_file.path)

    @property
    def results(self):
        """The results produced by the task. Set by get_results"""
        try:
            return self._results
        except AttributeError:
            self._results = self.get_results()
            return self._results 

    def get_results(self, *args, **kwargs):
        """
        Method called once the calculation is completed, 
        Updates self._results and returns TaskResults instance.
        Subclasses should extend this method (if needed) by adding 
        specialized code that performs some kind of post-processing.
        """
        # Check whether the process completed.
        if self.returncode is None:
            raise self.Error("return code is None, you should call wait, communitate or poll")

        if self.status != Task.S_DONE:
            raise self.Error("Task is not completed")

        return TaskResults({
            "task_name"      : self.name,
            "task_returncode": self.returncode,
            "task_status"    : self.status,
            "task_events"    : self.events.to_dict
        })

    def move(self, dst, isabspath=False):
        """
        Recursively move self.workdir to another location. This is similar to the Unix "mv" command.
        The destination path must not already exist. If the destination already exists
        but is not a directory, it may be overwritten depending on os.rename() semantics.

        Be default, dst is located in the parent directory of self.workdir, use isabspath=True
        to specify an absolute path.
        """
        if not isabspath:
            dst = os.path.join(os.path.dirname(self.workdir), dst)

        shutil.move(self.workdir, dst)


class AbinitTask(Task):
    """
    Base class defining an ABINIT calculation
    """
    # Prefixes for Abinit (input, output, temporary) files.
    Prefix = collections.namedtuple("Prefix", "idata odata tdata")
    pj = os.path.join

    prefix = Prefix(pj("indata", "in"), pj("outdata", "out"), pj("tmpdata", "tmp"))
    del Prefix, pj

    # Basenames for Abinit input and output files.
    Basename = collections.namedtuple("Basename", "input output files_file log_file stderr_file jobfile lockfile")
    basename = Basename("run.abi", "run.abo", "run.files", "run.log", "stderr", "job.sh", "__abilock__")
    del Basename

    def __init__(self, strategy, workdir, runmode, task_id=1, links=None, **kwargs):
        """
        Args:
            strategy: 
                `Strategy` instance describing the calculation.
            workdir:
                Path to the working directory.
            runmode:
                `RunMode` instance or string "sequential"
            task_id:
                Task identifier (must be unique if self belongs to a `Workflow`).
            links:
                List of `WorkLink` objects specifying the dependencies of the task.
                Used for tasks belonging to a `Workflow`.
            kwargs:
                keyword arguments (not used here)
        """
        super(AbinitTask, self).__init__()

        self.strategy = strategy

        self.workdir = os.path.abspath(workdir)

        self.runmode = RunMode.asrunmode(runmode).copy()

        self.set_id(task_id)

        # Connect this task to the other tasks 
        # (needed if are creating a Work instance with dependencies
        self._links = []
        if links is not None: 
            if not isinstance(links, collections.Iterable):
                links = [links,]

            self._links = links
            for link in links:
                print("Adding ", link.get_abivars())
                self.strategy.add_extra_abivars(link.get_abivars())

        # Files required for the execution.
        self.input_file = File(os.path.join(self.workdir, AbinitTask.basename.input))
        self.output_file = File(os.path.join(self.workdir, AbinitTask.basename.output))
        self.files_file = File(os.path.join(self.workdir, AbinitTask.basename.files_file))
        self.log_file = File(os.path.join(self.workdir, AbinitTask.basename.log_file))
        self.stderr_file = File(os.path.join(self.workdir, AbinitTask.basename.stderr_file))

        # Find number of processors ....
        #runhints = self.get_runhints()
        #launcher = self.runmode.make_launcher(runhints)

        # Build the launcher and store it.
        self.set_launcher(self.runmode.make_launcher(self))

    def __repr__(self):
        return "<%s at %s, workdir = %s>" % (
            self.__class__.__name__, id(self), os.path.basename(self.workdir))

    #def __str__(self):

    @classmethod
    def from_input(cls, abinit_input, workdir, runmode, task_id=1, links=None, **kwargs):
        """
        Create an instance of `AbinitTask` from an ABINIT input.

        Args:
            abinit_input:
                `AbinitInput` object.
            workdir:
                Path to the working directory.
            runmode:
                `RunMode` instance or string "sequential"
            task_id:
                Task identifier (must be unique if self belongs to a `Workflow`).
            links:
                List of `WorkLink` objects specifying the dependencies of the task.
                Used for tasks belonging to a `Workflow`.
            kwargs:
                keyword arguments (not used here)
        """
        # TODO: Find a better way to do this. I will likely need to refactor the Strategy object
        class StrategyWithInput(object):
            def __init__(self, abinit_input):
                self.abinit_input = abinit_input
                                                 
            @property
            def pseudos(self):
                return self.abinit_input.pseudos

            def add_extra_abivars(self, abivars):
                """Add variables (dict) to extra_abivars."""
                self.abinit_input.set_variables(**abivars)
                                                 
            def make_input(self):
                return str(self.abinit_input)

        strategy = StrategyWithInput(abinit_input)
        return cls(strategy, workdir, runmode, task_id=task_id, links=links, **kwargs)

    @property
    def name(self):
        return self.workdir

    @property
    def process(self):
        try:
            return self._process
        except AttributeError:
            # Attach a fake process so that we can poll it.
            return FakeProcess()

    @property
    def launcher(self):
        return self._launcher

    def set_launcher(self, launcher):
        self._launcher = launcher

    @property
    def jobfile_path(self):
        """Absolute path of the job file (shell script)."""
        return os.path.join(self.workdir, self.basename.jobfile)    

    @property
    def indata_dir(self):
        """Directory with the input data."""
        head, tail = os.path.split(self.prefix.idata)
        return os.path.join(self.workdir, head)

    @property
    def outdata_dir(self):
        """Directory with the output data."""
        head, tail = os.path.split(self.prefix.odata)
        return os.path.join(self.workdir, head)

    @property
    def tmpdata_dir(self):
        """Directory with the temporary data."""
        head, tail = os.path.split(self.prefix.tdata)
        return os.path.join(self.workdir, head)

    def idata_path_from_ext(self, ext):
        """Returns the path of the input file with extension ext."""
        return os.path.join(self.workdir, self.prefix.idata + ext)

    def odata_path_from_ext(self, ext):
        """Returns the path of the output file with extension ext"""
        return os.path.join(self.workdir, self.prefix.odata + ext)

    @property
    def pseudos(self):
        """List of pseudos used in the calculation."""
        return self.strategy.pseudos

    @property
    def filesfile_string(self):
        """String with the list of files and prefixex needed to execute ABINIT."""
        lines = []
        app = lines.append
        pj = os.path.join

        app(self.input_file.path)                 # Path to the input file
        app(self.output_file.path)                # Path to the output file
        app(pj(self.workdir, self.prefix.idata))  # Prefix for in data
        app(pj(self.workdir, self.prefix.odata))  # Prefix for out data
        app(pj(self.workdir, self.prefix.tdata))  # Prefix for tmp data

        # Paths to the pseudopotential files.
        for pseudo in self.pseudos:
            app(pseudo.path)

        return "\n".join(lines)

    @property
    def to_dict(self):
        raise NotImplementedError("")
        d = {k: v.to_dict for k, v in self.items()}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        d["strategy"] = self.strategy 
        d["workdir"] = workdir 
        return d
                                                                    
    @staticmethod
    def from_dict(d):
        raise NotImplementedError("")

    def isdone(self):
        """True if the output file is complete."""
        return abinit_output_iscomplete(self.output_file.path)

    @property
    def isnc(self):
        """True if norm-conserving calculation."""
        return all(p.isnc for p in self.pseudos)

    @property
    def ispaw(self):
        """True if PAW calculation"""
        return all(p.ispaw for p in self.pseudos)

    def make_input(self):
        """Construct and write the input file of the calculation."""
        return self.strategy.make_input()

    def outfiles(self):
        """Return all the output data files produced."""
        files = []
        for fname in os.listdir(self.outdata_dir):
            if fname.startswith(os.path.basename(self.prefix.odata)):
                files.append(os.path.join(self.outdata_dir, fname))

        return files
                                                                  
    def tmpfiles(self):
        """Return all the input data files produced."""
        files = []
        for fname in os.listdir(self.tmpdata_dir):
            if file.startswith(os.path.basename(self.prefix.tdata)):
                files.append(os.path.join(self.tmpdata_dir, fname))

        return files

    def path_in_workdir(self, filename):
        """Create the absolute path of filename in the top-level working directory."""
        return os.path.join(self.workdir, filename)

    def path_in_indatadir(self, filename):
        """Create the absolute path of filename in the indata directory."""
        return os.path.join(self.indata_dir, filename)

    def path_in_outdatadir(self, filename):
        """Create the absolute path of filename in the outdata directory."""
        return os.path.join(self.outdata_dir, filename)

    def path_in_tmpdatadir(self, filename):
        """Create the absolute path of filename in the tmp directory."""
        return os.path.join(self.tmpdata_dir, filename)

    def rename(self, src_basename, dest_basename, datadir="outdir"):
        """
        Rename a file located in datadir.

        src_basename and dest_basename are the basename of the source file
        and of the destination file, respectively.
        """
        if datadir == "outdir":
            src = self.path_in_outdatadir(src_basename)
            dest = self.path_in_outdatadir(dest_basename)

        elif datadir == "tmpdir":
            src = self.path_in_tmpdatadir(src_basename)
            dest = self.path_in_tmpdatadir(dest_basename)

        else:
            raise ValueError("Wrong datadir %s" % datadir)

        os.rename(src, dest)

    def build(self, *args, **kwargs):
        """
        Creates the working directory and the input files of the `Task`.
        It does not overwrite files if they already exist.
        """
        # Top-level dir
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

        # Create dirs for input, output and tmp data.
        if not os.path.exists(self.indata_dir):
            os.makedirs(self.indata_dir)

        if not os.path.exists(self.outdata_dir):
            os.makedirs(self.outdata_dir)

        if not os.path.exists(self.tmpdata_dir):
            os.makedirs(self.tmpdata_dir)

        # Write input file and files file.
        if not self.input_file.exists:
            self.input_file.write(self.make_input())

        if not self.files_file.exists:
            self.files_file.write(self.filesfile_string)

        #if not os.path.exists(self.jobfile_path):
        #    with open(self.jobfile_path, "w") as f:
        #            f.write(str(self.jobfile))

    def rmtree(self, exclude_wildcard=""):
        """
        Remove all files and directories in the working directory

        Args:
            exclude_wildcard:
                Optional string with regular expressions separated by |.
                Files matching one of the regular expressions will be preserved.
                example: exclude_wildcard="*.nc|*.txt" preserves all the files
                whose extension is in ["nc", "txt"].
        """
        if not exclude_wildcard:
            shutil.rmtree(self.workdir)
        else:
            pats = exclude_wildcard.split("|")

            import fnmatch

            def keep(fname):
                for pat in pats:
                    if fnmatch.fnmatch(fname, pat):
                        return True
                return False

            for dirpath, dirnames, filenames in os.walk(self.workdir):
                for fname in filenames:
                    filepath = os.path.join(dirpath, fname)
                    if not keep(fname):
                        os.remove(filepath)

    def remove_files(self, *filenames):
        """Remove all the files listed in filenames."""
        if isinstance(filenames, str):
            filenames = [filenames]

        for dirpath, dirnames, fnames in os.walk(self.workdir):
            for fname in fnames:
                if fname in filenames:
                    filepath = os.path.join(dirpath, fname)
                    os.remove(filepath)

    def rm_tmpdatadir(self):
        """Remove the directory with the temporary files."""
        shutil.rmtree(self.tmpdata_dir)

    def rm_indatadir(self):
        """Remove the directory with the input files (indata dir)."""
        shutil.rmtree(self.indata_dir)

    def get_runhints(self):
        """
        Run ABINIT in sequential to obtain a set of possible
        configurations for the number of processors.
        RunMode is used to select the paramenters of the run.
        """
        raise NotImplementedError("")

        # Build a sequential launcher to run the code.
        seq_launcher = self.runmode.make_seq_launcher(self)

        # Launch the calculation.
        retcode = seq_launcher.launch(self)

        # Parse the output file to find the number of MPI processors, 
        # the number of OpenMP processors, memory estimate in Gb ...
        runhints = RunHintsParser().parse(self.output_file.path)

        self.set_status(self.S_READY)

        return runhints

    def setup(self, *args, **kwargs):
        pass

    def start(self, *args, **kwargs):
        """
        Starts the calculation by performing the following steps:

            - build dirs and files
            - call the _setup method
            - execute the job file via the shell or other means.
            - return Results object

        .. warning::
             This method must be thread safe since we may want to run several independent
             calculations with different python threads. 
        """
        self.build(*args, **kwargs)

        self._setup(*args, **kwargs)

        # Start the calculation in a subprocess and return.
        self._process = self.launcher.launch(self)

    def start_and_wait(self, *args, **kwargs):
        """Helper method to start the task and wait."""
        self.start(*args, **kwargs)
        return self.wait()

##########################################################################################

class TaskResults(dict, MSONable):
    """
    Dictionary used to store the most important results produced by a Task.
    """
    _mandatory_keys = [
        "task_name",
        "task_returncode",
        "task_status",
        "task_events",
    ]
    EXC_KEY = "_exceptions"

    def __init__(self, *args, **kwargs):
        super(TaskResults, self).__init__(*args, **kwargs)
                                                               
        if self.EXC_KEY not in self:
            self[self.EXC_KEY] = []

    @property
    def exceptions(self):
        return self[self.EXC_KEY]

    def push_exceptions(self, *exceptions):
        for exc in exceptions:
            newstr = str(exc)
            if newstr not in self.exceptions:
                self[self.EXC_KEY] += [newstr,]

    def assert_valid(self):
        """
        Returns an empty list if results seem valid. 

        The try assert except trick allows one to get a string with info on the exception.
        We use the += operator so that sub-classes can add their own message.
        """
        # TODO Better treatment of events.
        try:
            assert (self["task_returncode"] == 0 and self["task_status"] == Task.S_DONE)
        except AssertionError as exc:
            self.push_exceptions(exc)

        return self.exceptions

    @property
    def to_dict(self):
        d = {k: v for k,v in self.items()}
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        return d
                                                                                
    @classmethod
    def from_dict(cls, d):
        return cls({k: v for k,v in d.items() if k not in ["@module", "@class",]})

    def json_dump(self, filename):
        json_pretty_dump(self.to_dict, filename) 

    @classmethod
    def json_load(cls, filename):
        return cls.from_dict(json_load(filename))    


class RunHints(collections.OrderedDict):
    """
    Dictionary with the hints for the parallel execution reported by abinit.

    Example:
        <RUN_HINTS, max_ncpus = "108", autoparal="3">
            "1" : {"tot_ncpus": 2,           # Total number of CPUs
                   "mpi_ncpus": 2,           # Number of MPI processes.
                   "omp_ncpus": 1,           # Number of OMP threads (0 if not used)
                   "memory_gb": 10,          # Estimated memory requirement in Gigabytes (total or per proc?)
                   "weight"   : 0.4,         # 1.0 corresponds to an "expected" optimal efficiency.
                   "variables": {            # Dictionary with the variables that should be added to the input.
                        "varname1": varvalue1,
                        "varname2": varvalue2,
                        }
                }
        </RUN_HINTS>

    For paral_kgb we have:
    nproc     npkpt  npspinor    npband     npfft    bandpp    weight   
       108       1         1        12         9         2        0.25
       108       1         1       108         1         2       27.00
        96       1         1        24         4         1        1.50
        84       1         1        12         7         2        0.25
    """
    @classmethod
    def from_file(cls, filename):
        """
        Read the <RUN_HINTS> section from file filename
        Assumes the file contains only one section.
        """
        with open(filaname, "r") as fh:
            lines = fh.readlines()

        try:
            start = lines.index("<RUN_HINTS>\n")
            stop  = lines.index("</RUN_HINTS>\n")
        except ValueError:
            raise ValueError("%s does not contain any valid RUN_HINTS section" % filename)

        runhints = json_loads("".join(l for l in lines[start+1:stop]))

        # Sort the dict so that configuration with minimum (weight-1.0) appears in the first positions
        return cls(sorted(runhints, key=lambda t : abs(t[1]["weight"] - 1.0), reverse=False))

    def __init__(self, *args, **kwargs):
        super(RunHints, self).__init__(*args, **kwargs)

    def get_hint(self, policy):
        # Find the optimal configuration depending on policy
        raise NotImplementedError("")
        # Copy the dictionary and return AttrDict instance.
        #odict = self[okey].copy()
        #odict["_policy"] = policy
        #return AttrDict(odict)


class RunMode(dict, MSONable):
    """
    This object contains the user-specified settings controlling the execution 
    of the run (submission, coarse- and fine-grained parallelism ...)
    It acts as a factory that produced Launcher instances.
    """
    # Default values (correspond to the sequential mode).
    _defaults = {
        "launcher_type": "shell",    # ["shell", "slurm", "slurm", "pbs",]
        #"with_fw"     : False,      # True if we are using fireworks.
        "policy"       : "default",  # Policy used to select the number of MPI processes when there are ambiguities.
        "mpirun"       : "mpirun",   # Path to the mpi runnner.
        "max_ncpus"    : np.inf,     # Max number of CPUs that can be used (DEFAULT: no limit is enforces)
        "omp_env"      : {},         # OpenMP environment variables.
        "queue_params" : {},         # Parameters passed to the firework QueueAdapter.
    }

    def __init__(self, *args, **kwargs):
        super(RunMode, self).__init__(*args, **kwargs)

    def copy(self):
        return RunMode(**super(RunMode, self).copy())
    
    @classmethod
    def asrunmode(cls, obj):
        """
        Returns a RunMode instance from obj. Accepts:

            - RunMode instances
            - A string with "sequential" to indicate that Abinit will be executed in sequential mode.
        """
        if isinstance(obj, cls):
            return obj

        if isinstance(obj, str) and obj == "sequential":
            return cls.sequential()

        raise ValueError("Don't know how to convert obj %s into a %s instance" % (obj, obj.__class__.__name__))

    @classmethod
    def load_user_configuration(cls):
        fname = "runmode.json"
        cwd = os.getcwd()

        path = os.path.join(cwd, fname)

        if os.path.exists(path):
            return cls.from_filename(path)
        else:
            raise NotImplementedError("Trying in the pymatgen dir")
            #raise RuntimeError("Cannot find configuration file for initializing RunMode instance")

    @classmethod
    def from_file(cls, filename):
        """Initialize an instance of RunMode from the configuration file (JSON format)."""
        defaults = cls._defaults.copy() 
        d = json_load(filename) 

        # Put default values if they are not in d.
        for (k,v) in defaults.items():
            if k not in d:
                d[k] = v
        return cls(d)

    @classmethod
    def sequential(cls, launcher_type=None):
        d = cls._defaults.copy()
        if launcher_type is not None:
            d["launcher_type"] = launcher_type

        return cls(d)

    @classmethod
    def mpi_parallel(cls, mpi_ncpus, launcher_type=None):
        d = cls._defaults.copy()
        if launcher_type is not None:
            d["launcher_type"] = launcher_type
        new = cls(d)
        new.set_mpi_ncpus(mpi_ncpus)
        return new

    def set_mpi_ncpus(self, mpi_ncpus):
        self["mpi_ncpus"] = min(mpi_ncpus, self["max_ncpus"])

    #def set_omp_ncpus(self, omp_ncpus):
    #     self["omp_ncpus"] = min(omp_ncpus, self["max_ncpus"])

    @property
    def to_dict(self):
        d = self.copy()
        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__
        return d
                                                 
    @classmethod
    def from_dict(cls, d):
        return cls({k: v for (k, v) in d.items() if k not in ("@module", "@class")})

    def make_launcher(self, task):
        """Factory function returning a launcher instance."""
        #if have_fw:
        #   raise ValueError("Firework not yet supported")
                                                       
        # Find the subclass to instanciate.
        cls = TaskLauncher.from_launcher_type(task.runmode["launcher_type"])

        return cls(task.jobfile_path, 
                   stdin  = task.files_file.path, 
                   stdout = task.log_file.path,
                   stderr = task.stderr_file.path,
                   **task.runmode
                  )

    def make_seq_launcher(self, task):
        """Return a simple launcher for sequential runs."""
        raise NotImplementedError("")

##########################################################################################
